import os
import sys

sys.path.append(".")
from mlc.logging import XP

import torch
import optuna
from optuna.trial import TrialState
from optuna._callbacks import MaxTrialsCallback
from mlc.datasets.dataset_factory import load_dataset
from mlc.models.model_factory import load_model
from mlc.utils import cross_validation
from mlc.load_do_save import save_json
from mlc.utils import parent_exists
from argparse import Namespace, ArgumentParser
from mlc.transformers.tab_scaler import TabScaler
from mlc.metrics.compute import compute_metric
from mlc.metrics.metric_factory import create_metric
from mlc.dataloaders import get_custom_dataloader

# Torch config to avoid crash on HPC
torch.multiprocessing.set_sharing_strategy('file_system')



class Objective(object):
    def __init__(self, args, model_class, scaler, X, y):
        # Save the model that will be trained
        self.model_class = model_class

        # Save the trainings data
        self.X = X
        self.y = y
        self.scaler = scaler

        self.args = args

    def __call__(self, trial):
        # Define hyperparameters to optimize
        trial_params = self.model_class.define_trial_parameters(trial, self.args)

        # Create model
        args = {**self.args, **trial_params, "task": "HP train", "trial": trial.number}

        model = self.model_class(**args, scaler=self.scaler)
        args = Namespace(**args)

        # Cross validate the chosen hyperparameters
        metric = args.metrics[0]
        metrics, time = cross_validation(model, self.X, self.y, metric, args)

        experiment = XP(vars(args))
        save_path = f"./tmp/models/{args.dataset}_{args.model_name}.model"
        model.save(save_path)
        experiment.log_model(f"{args.dataset}_{args.model_name}.model", save_path)
        experiment.log_metric(metric, metrics.mean())
        experiment.end()

        return metrics.mean()


def run(dataset_name: str = "lcld_v2_iid", model_name: str = "tabtransformer", n_trials: int = 100, subset: int = 1000,
        train_batch_size: int = 256, val_batch_size: int = 1024, save_best_model: int = 0, epochs: int = 100,
        custom_dataloader:str="") -> None:
    print("Train hyperparameter optimization for {} on {}".format(model_name, dataset_name))
    dataset = load_dataset(dataset_name)
    metadata = dataset.get_metadata(only_x=True)
    common_model_params = {
        "x_metadata": metadata,
        "objective": "classification",
        "use_gpu": True,
        "batch_size": train_batch_size,
        "num_classes": 2,
        "early_stopping_rounds": 5,
        "val_batch_size": val_batch_size,
        "class_weight": "balanced",
        "custom_dataloader":custom_dataloader
    }

    x, y = dataset.get_x_y()
    splits = dataset.get_splits()

    x_train = x.iloc[splits["train"]]
    y_train = y[splits["train"]]

    if subset > 0:
        x_train = x_train[:subset]
        y_train = y_train[:subset]

    model_class = load_model(model_name)
    args = {
        **common_model_params,
        "depth": 12,
        "heads": 2,  # 4,
        "weight_decay": -3,  # -2,
        "learning_rate": -3,  # -2,
        "dropout": 0.5,  # 0.3,
        "data_parallel": False,
        "epochs": epochs,
        "model_name": model_name,
        "dataset": dataset_name,
        "early_stopping_rounds": 5,
        "num_splits": 5,
        "seed": 42,
        "shuffle": True,
        "metrics": ["auc"]
    }
    if model_name == "torchrln":
        args["weight_decay"] = 0

    study_name = model_name + "_" + dataset_name
    storage_name = "sqlite:///{}.db".format(study_name)

    study = optuna.create_study(direction="maximize",
                                study_name=study_name,
                                storage=storage_name,
                                load_if_exists=True)

    n_completed = len(study.get_trials(states=(TrialState.COMPLETE,)))
    n_to_finish = n_trials
    print(f"{n_completed}/{n_trials} already completed")

    scaler = TabScaler(num_scaler="min_max", one_hot_encode=True)
    scaler.fit(
        torch.tensor(x.values, dtype=torch.float32), x_type=metadata["type"]
    )

    if n_completed < n_to_finish:
        study.optimize(
            Objective(args, model_class, scaler, x_train.to_numpy(), y_train),
            n_trials=None,
            callbacks=[
                MaxTrialsCallback(
                    n_to_finish,
                    states=(TrialState.COMPLETE,),
                ),
            ],
        )

    save_path = os.path.join(".", "data", "models", "{}.model".format(model_name), "best_args.json")
    best_args = {**args, **study.best_trial.params, "task": "HP train", "trial": "best"}
    best_args["x_metadata"] = None
    experiment = XP(best_args)
    experiment.log_metrics({"best_{}".format(args.get("metrics")[0]): study.best_value})

    save_json(best_args, parent_exists(save_path))
    print(f"Best parameters: {best_args} with best performance: {study.best_value}")

    if save_best_model:
        best_trial = sorted(study.trials,key=lambda d: d.value if d.values is not None else -1)[-1]
        print(f"Best trial parameters: {best_trial.params} with best performance: {best_trial.value}")
        model = model_class(**{**args,**best_trial.params, "early_stopping_rounds":args["epochs"]}, scaler=scaler)
        x_test = x.iloc[splits["test"]]
        y_test = y[splits["test"]]

        custom_train_dataloader = get_custom_dataloader(custom_dataloader, dataset,model, scaler,{},
                                                        x=x_train.values,y=y_train, train=True, batch_size=train_batch_size)
        model.fit(x_train,y_train,x_test,y_test, custom_train_dataloader=custom_train_dataloader)

        sc = create_metric("auc")
        metric_val = compute_metric(model, sc, x_test, y_test)
        print(metric_val)
        experiment.log_metric("test_{}".format(sc.metric_name), metric_val)

        save_path = os.path.join(".", "data", "models", "best_{}_{}_{}.model".format(model_name,dataset_name,
                                                                                     custom_dataloader))
        model.save(save_path)
        experiment.log_model("best_{}_{}_{}.model".format(model_name,dataset_name,custom_dataloader),save_path)

if __name__ == "__main__":
    parser = ArgumentParser(
        description="Training with Hyper-parameter optimization"
    )
    parser.add_argument("--dataset_name", type=str, default="lcld_v2_iid",
                        )
    parser.add_argument("--model_name", type=str, default="tabtransformer",
                        )
    parser.add_argument("--n_trials", type=int, default=100)
    parser.add_argument("--epochs", type=int, default=100)
    parser.add_argument("--subset", type=int, default=0)
    parser.add_argument("--train_batch_size", type=int, default=1024)
    parser.add_argument("--val_batch_size", type=int, default=2048)
    parser.add_argument("--save_best_model", type=int, default=1)
    parser.add_argument("--custom_dataloader", type=str, default="")

    args = parser.parse_args()

    run(dataset_name=args.dataset_name, model_name=args.model_name, n_trials=args.n_trials, subset=args.subset,
        train_batch_size=args.train_batch_size, val_batch_size=args.val_batch_size,
        save_best_model=args.save_best_model,
        epochs=args.epochs,custom_dataloader=args.custom_dataloader)



from __future__ import annotations

import abc
import os
from abc import ABC
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
import requests
from sklearn.model_selection import train_test_split

from mlc.constraints.constraints import (
    Constraints,
    get_constraints_from_metadata,
)
from mlc.constraints.constraints_checker import ConstraintChecker
from mlc.constraints.relation_constraint import BaseRelationConstraint


class SplitterMissingError(Exception):
    pass


class DataSource(ABC):
    @abc.abstractmethod
    def load_data(self) -> pd.DataFrame:
        pass


class FileDataSource(DataSource, ABC):
    def __init__(self, path: str):
        self.path = path

    def get_path(self) -> str:
        return self.path


class CsvDataSource(FileDataSource):
    def __init__(self, path: str, **kwargs: Dict[str, Any]) -> None:
        self.pandas_params = kwargs
        super(CsvDataSource, self).__init__(path)

    def load_data(self) -> pd.DataFrame:
        return pd.read_csv(self.path, **self.pandas_params, low_memory=False)


class Hdf5DataSource(FileDataSource):
    def __init__(self, path: str, **kwargs: Dict[str, Any]) -> None:
        self.pandas_params = kwargs
        super().__init__(path)

    def load_data(self) -> pd.DataFrame:
        data = pd.read_hdf(self.path, **self.pandas_params, low_memory=False)
        if isinstance(data, pd.DataFrame):
            return data
        else:
            raise NotImplementedError


class SQLDataSource(DataSource):
    def __init__(
        self,
        connection_string: str,
        query: str,
        preprocess_fun: Optional[
            Callable[[pd.DataFrame], pd.DataFrame]
        ] = None,
    ) -> None:
        self.connection_string = connection_string
        self.query = query
        self.preprocess_fun = preprocess_fun

    def load_data(self) -> pd.DataFrame:
        df = pd.read_sql(self.query, self.connection_string)
        if self.preprocess_fun is not None:
            df = self.preprocess_fun(df)
        return df


class DownloadFileDataSource(DataSource):
    def __init__(
        self,
        url: str,
        file_data_source: FileDataSource,
        overwrite: bool = False,
    ) -> None:
        self.url = url
        self.file_data_source = file_data_source
        self.overwrite = overwrite

    def load_data(self) -> pd.DataFrame:
        path = self.file_data_source.get_path()
        if self.overwrite or (not os.path.exists(path)):
            Path(path).parent.mkdir(parents=True, exist_ok=True)
            data = requests.get(self.url)
            with open(path, "wb") as file:
                file.write(data.content)
        return self.file_data_source.load_data()


class TaskProcessor(ABC):
    @abc.abstractmethod
    def transform(self, task_data: pd.Series) -> pd.Series:
        pass


class Task:
    def __init__(
        self,
        name: Union[int, str],
        task_type: str,
        evaluation_metric: str,
        task_processor: TaskProcessor = None,
    ):
        self.name = name
        self.type = task_type
        self.evaluation_metric = evaluation_metric
        self.task_processor = task_processor


class Sorter(ABC):
    @abc.abstractmethod
    def get_index(
        self, data: pd.DataFrame
    ) -> Union[pd.Series, npt.NDArray[np.int_]]:
        pass


class DefaultIndexSorter(Sorter):
    def get_index(
        self, data: pd.DataFrame
    ) -> Union[pd.Series, npt.NDArray[np.int_]]:
        return np.arange(len(data))


class DateSorter(Sorter):
    def __init__(self, date_col: str) -> None:
        self.date_col = date_col

    def get_index(
        self, data: pd.DataFrame
    ) -> Union[pd.Series, npt.NDArray[np.int_]]:
        return pd.to_datetime(data[self.date_col])


class Splitter(ABC):
    @abc.abstractmethod
    def get_splits(self, dataset: Dataset) -> Dict[str, npt.NDArray[np.int_]]:
        pass


class DefaultSplitter(Splitter):
    def __init__(
        self, test_size: Union[float, int] = 0.2, random_seed: int = 42
    ) -> None:
        self.test_size = test_size
        self.random_seed = random_seed

    def get_splits(self, dataset: Dataset) -> Dict[str, npt.NDArray[np.int_]]:
        _, y = dataset.get_x_y()
        i = np.arange(len(y))
        i_train, i_test = train_test_split(
            i,
            random_state=self.random_seed,
            shuffle=True,
            stratify=y[i],
            test_size=self.test_size,
        )
        i_train, i_val = train_test_split(
            i_train,
            random_state=self.random_seed,
            shuffle=True,
            stratify=y[i_train],
            test_size=self.test_size,
        )
        return {"train": i_train, "val": i_val, "test": i_test}


class Dataset:
    def __init__(
        self,
        name: str,
        data_source: DataSource,
        metadata_source: DataSource,
        tasks: List[Task],
        sorter: Sorter,
        splitter: Union[Splitter, None],
        relation_constraints: List[BaseRelationConstraint] = None,
    ):

        self.name = name
        self.data_source = data_source
        self.metadata_source = metadata_source
        self.splitter = splitter
        self.tasks = tasks
        self.sorter = sorter
        self.relation_constraints = relation_constraints

        self.data: Optional[pd.DataFrame] = None
        self.metadata: Optional[pd.DataFrame] = None

    def get_name(self) -> str:
        return self.name

    def set_data(self, df: pd.DataFrame, filter_constraints: bool = False):
        data = df.copy()
        if filter_constraints:
            checker = ConstraintChecker(self.get_constraints(), tolerance=0.01)
            check = checker.check_constraints(self.data, data)
            data = data[check]
        self.data = data

    def get_data(self) -> pd.DataFrame:
        if self.data is None:
            self.data = self.data_source.load_data()
        return self.data.copy()

    def get_metadata(self, only_x: bool = False) -> pd.DataFrame:
        if self.metadata is None:
            self.metadata = self.metadata_source.load_data()
        metadata = self.metadata.copy()
        if only_x:
            x_col = self.get_x_y()[0].columns
            metadata = metadata[metadata["feature"].isin(x_col)]

        return metadata

    def get_ddpm_transformations(self):

        return {
            "seed": 0,
            "normalization": "quantile",
            "num_nan_policy": "__none__",
            "cat_nan_policy": "__none__",
            "cat_min_frequency": "__none__",
            "cat_encoding": "__none__",
            "y_policy": "default",
        }

    def get_x_y(
        self, keep_date: bool = False
    ) -> Tuple[
        pd.DataFrame, Union[npt.NDArray[np.int_], npt.NDArray[np.float_]]
    ]:
        data = self.get_data()
        y = []
        for task in self.tasks:
            task_data = data[task.name]
            if task.task_processor is not None:
                task_data = task.task_processor.transform(task_data)
            data = data.drop(columns=task.name)
            y.append(task_data)

        y_np = np.array(y)
        if y_np.shape[0] == 1:
            y_np = y_np.ravel()

        if isinstance(self.sorter, DateSorter) and (not keep_date):
            data = data.drop(columns=self.sorter.date_col)

        return data, y_np

    def get_x_y_t(
        self, keep_date: bool = False
    ) -> Tuple[
        pd.DataFrame,
        Union[npt.NDArray[np.int_], npt.NDArray[np.float_]],
        Union[pd.Series, npt.NDArray[np.int_]],
    ]:
        x, y = self.get_x_y(keep_date=True)
        t = self.sorter.get_index(x)
        if isinstance(self.sorter, DateSorter) and (not keep_date):
            x = x.drop(columns=self.sorter.date_col)
        return x, y, t

    def get_splits(self) -> Dict[str, npt.NDArray[np.int_]]:
        if self.splitter is None:
            raise SplitterMissingError
        return self.splitter.get_splits(self)

    def get_constraints(self) -> Constraints:
        metadata = self.get_metadata()
        col_filter = self.get_x_y()[0].columns.to_list()
        relation_constraints = self.relation_constraints
        if relation_constraints is None:
            relation_constraints = []
        return get_constraints_from_metadata(
            metadata, relation_constraints, col_filter
        )



import numpy as np
import torch
import zero

from .gaussian_multinomial_diffusion import GaussianMultinomialDiffusion
from .lib import Transformations
from .utils_train import get_model, make_dataset


def to_good_ohe(ohe, X):
    indices = np.cumsum([0] + ohe._n_features_outs)
    Xres = []
    for i in range(1, len(indices)):
        x_ = np.max(X[:, indices[i - 1]:indices[i]], axis=1)
        t = X[:, indices[i - 1]:indices[i]] - x_.reshape(-1, 1)
        Xres.append(np.where(t >= 0, 1, 0))
    return np.hstack(Xres)


def sample(
        parent_dir,
        real_data_path='data/higgs-small',
        batch_size=2000,
        num_samples=0,
        model_type='mlp',
        model_params=None,
        model=None,
        model_path=None,
        num_timesteps=1000,
        gaussian_loss_type='mse',
        scheduler='cosine',
        T_dict=None,
        num_numerical_features=0,
        disbalance=None,
        device=torch.device('cuda:1'),
        seed=0,
        change_val=False
):
    zero.improve_reproducibility(seed)

    T = Transformations(**T_dict)
    D = make_dataset(
        real_data_path,
        T,
        num_classes=model_params['num_classes'],
        is_y_cond=model_params['is_y_cond'],
        change_val=change_val
    )

    K = np.array(D.get_category_sizes('train'))
    if len(K) == 0 or T_dict['cat_encoding'] == 'one-hot':
        K = np.array([0])

    num_numerical_features_ = D.X_num['train'].shape[1] if D.X_num is not None else 0
    d_in = np.sum(K) + num_numerical_features_
    model_params['d_in'] = int(d_in)

    if model is None:
        model = get_model(
            model_type,
            model_params,
            num_numerical_features_,
            category_sizes=D.get_category_sizes('train')
        )

        model.load_state_dict(
            torch.load(model_path, map_location="cpu")
        )

    diffusion = GaussianMultinomialDiffusion(
        K,
        num_numerical_features=num_numerical_features_,
        denoise_fn=model, num_timesteps=num_timesteps,
        gaussian_loss_type=gaussian_loss_type, scheduler=scheduler, device=device
    )

    diffusion.to(device)
    diffusion.eval()

    _, empirical_class_dist = torch.unique(torch.from_numpy(D.y['train']), return_counts=True)
    # empirical_class_dist = empirical_class_dist.float() + torch.tensor([-5000., 10000.]).float()
    if disbalance == 'fix':
        empirical_class_dist[0], empirical_class_dist[1] = empirical_class_dist[1], empirical_class_dist[0]
        x_gen, y_gen = diffusion.sample_all(num_samples, batch_size, empirical_class_dist.float(), ddim=False)

    elif disbalance == 'fill':
        ix_major = empirical_class_dist.argmax().item()
        val_major = empirical_class_dist[ix_major].item()
        x_gen, y_gen = [], []
        for i in range(empirical_class_dist.shape[0]):
            if i == ix_major:
                continue
            distrib = torch.zeros_like(empirical_class_dist)
            distrib[i] = 1
            num_samples = val_major - empirical_class_dist[i].item()
            x_temp, y_temp = diffusion.sample_all(num_samples, batch_size, distrib.float(), ddim=False)
            x_gen.append(x_temp)
            y_gen.append(y_temp)

        x_gen = torch.cat(x_gen, dim=0)
        y_gen = torch.cat(y_gen, dim=0)

    else:
        x_gen, y_gen = diffusion.sample_all(num_samples, batch_size, empirical_class_dist.float(), ddim=False)

    # try:
    # except FoundNANsError as ex:
    #     print("Found NaNs during sampling!")
    #     loader = lib.prepare_fast_dataloader(D, 'train', 8)
    #     x_gen = next(loader)[0]
    #     y_gen = torch.multinomial(
    #         empirical_class_dist.float(),
    #         num_samples=8,
    #         replacement=True
    #     )
    X_gen, y_gen = x_gen.numpy(), y_gen.numpy()

    return X_gen, y_gen



from typing import List, Optional, Union

import numpy as np
import torch

from mlc.constraints.constraints_backend_executor import ConstraintsExecutor
from mlc.constraints.pytorch_backend import PytorchBackend
from mlc.constraints.relation_constraint import (
    BaseRelationConstraint,
    EqualConstraint,
    Feature,
)
from mlc.constraints.utils import get_feature_index
from mlc.typing import NDFloat
from mlc.utils import to_numpy_number, to_torch_number


class ConstraintsFixer:
    def __init__(
        self,
        guard_constraints: List[BaseRelationConstraint],
        fix_constraints: List[EqualConstraint],
        feature_names: Optional[List[str]] = None,
    ):
        self.guard_constraints = guard_constraints
        self.fix_constraints = fix_constraints
        self.feature_names = feature_names
        for c in self.fix_constraints:
            if not isinstance(c, EqualConstraint):
                raise ValueError("Fix constraints must be an EqualConstraint.")
            else:
                if not isinstance(c.left_operand, Feature):
                    raise ValueError(
                        "Left operand of fix constraints must be a Feature."
                    )

    def fix(
        self, x: Union[torch.Tensor, NDFloat]
    ) -> Union[torch.Tensor, NDFloat]:

        if isinstance(x, np.ndarray):
            return to_numpy_number(self.fix(to_torch_number(x)))

        x = x.clone()

        for i in range(len(self.fix_constraints)):
            guard_c = self.guard_constraints[i]
            fix_c = self.fix_constraints[i]

            if not isinstance(fix_c.left_operand, Feature):
                raise ValueError(
                    "Left operand of fix constraints must be a Feature."
                )

            # Index of inputs that shall be updated
            # according the guard constraints,
            # if none then update all.
            if guard_c is not None:
                executor = ConstraintsExecutor(
                    guard_c, PytorchBackend(), self.feature_names
                )
                to_update = executor.execute(x) > 0
            else:
                to_update = torch.ones(x.shape[0], dtype=torch.bool)

            # Index of the feature to update.
            # Ignore warning, this is checked in the constructor.
            index = get_feature_index(
                self.feature_names, fix_c.left_operand.feature_id
            )

            # Value to be update.
            # Known warning, not supposed to evaluate without constraint.
            executor = ConstraintsExecutor(
                fix_c.right_operand, PytorchBackend(), self.feature_names
            )
            new_value = executor.execute(x[to_update])

            x[to_update, index] = new_value

        return x



from .gaussian_multinomial_diffusion import GaussianMultinomialDiffusion
from .lib import *
from .utils_train import get_model, make_dataset, update_ema


class Trainer:
    def __init__(self, diffusion, train_iter, lr, weight_decay, steps, device=torch.device('cuda:1')):
        self.diffusion = diffusion
        self.ema_model = deepcopy(self.diffusion._denoise_fn)
        for param in self.ema_model.parameters():
            param.detach_()

        self.train_iter = train_iter
        self.steps = steps
        self.init_lr = lr
        self.optimizer = torch.optim.AdamW(self.diffusion.parameters(), lr=lr, weight_decay=weight_decay)
        self.device = device
        self.loss_history = pd.DataFrame(columns=['step', 'mloss', 'gloss', 'loss'])
        self.log_every = 100
        self.print_every = 500
        self.ema_every = 1000

    def _anneal_lr(self, step):
        frac_done = step / self.steps
        lr = self.init_lr * (1 - frac_done)
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = lr

    def _run_step(self, x, out_dict):
        x = x.to(self.device)
        for k in out_dict:
            out_dict[k] = out_dict[k].long().to(self.device)
        self.optimizer.zero_grad()
        loss_multi, loss_gauss = self.diffusion.mixed_loss(x, out_dict)
        loss = loss_multi + loss_gauss
        loss.backward()
        self.optimizer.step()

        return loss_multi, loss_gauss

    def run_loop(self):
        step = 0
        curr_loss_multi = 0.0
        curr_loss_gauss = 0.0

        curr_count = 0
        while step < self.steps:
            x, out_dict = next(self.train_iter)
            out_dict = {'y': out_dict}
            batch_loss_multi, batch_loss_gauss = self._run_step(x, out_dict)

            self._anneal_lr(step)

            curr_count += len(x)
            curr_loss_multi += batch_loss_multi.item() * len(x)
            curr_loss_gauss += batch_loss_gauss.item() * len(x)

            if (step + 1) % self.log_every == 0:
                mloss = np.around(curr_loss_multi / curr_count, 4)
                gloss = np.around(curr_loss_gauss / curr_count, 4)
                if (step + 1) % self.print_every == 0:
                    print(f'Step {(step + 1)}/{self.steps} MLoss: {mloss} GLoss: {gloss} Sum: {mloss + gloss}')
                self.loss_history.loc[len(self.loss_history)] = [step + 1, mloss, gloss, mloss + gloss]
                curr_count = 0
                curr_loss_gauss = 0.0
                curr_loss_multi = 0.0

            update_ema(self.ema_model.parameters(), self.diffusion._denoise_fn.parameters())

            step += 1


def train(
        parent_dir,
        dataset=None,
        real_data_path='data/higgs-small',
        steps=1000,
        lr=0.002,
        weight_decay=1e-4,
        batch_size=1024,
        model_type='mlp',
        model_params={"is_y_cond": True, "num_classes": 2, "rtdl_params": {"d_layers": [256, 256], "dropout": 0.0}},
        num_timesteps=1000,
        gaussian_loss_type='mse',
        scheduler='cosine',
        T_dict={},
        num_numerical_features=0,
        device=torch.device('cuda:0'),
        seed=0,
        change_val=False
):
    real_data_path = os.path.normpath(real_data_path)
    parent_dir = os.path.normpath(parent_dir)

    zero.improve_reproducibility(seed)

    T = Transformations(**T_dict)

    if dataset is None:
        dataset = make_dataset(
            real_data_path,
            T,
            num_classes=model_params['num_classes'],
            is_y_cond=model_params['is_y_cond'],
            change_val=change_val
        )

    K = np.array(dataset.get_category_sizes('train'))
    if len(K) == 0 or T_dict['cat_encoding'] == 'one-hot':
        K = np.array([0])
    print(K)

    num_numerical_features = dataset.X_num['train'].shape[1] if dataset.X_num is not None else 0
    d_in = np.sum(K) + num_numerical_features
    model_params['d_in'] = d_in
    print(d_in)

    print(model_params)
    model = get_model(
        model_type,
        model_params,
        num_numerical_features,
        category_sizes=dataset.get_category_sizes('train')
    )
    model.to(device)

    # train_loader = lib.prepare_beton_loader(dataset, split='train', batch_size=batch_size)
    train_loader = prepare_fast_dataloader(dataset, split='train', batch_size=batch_size)

    diffusion = GaussianMultinomialDiffusion(
        num_classes=K,
        num_numerical_features=num_numerical_features,
        denoise_fn=model,
        gaussian_loss_type=gaussian_loss_type,
        num_timesteps=num_timesteps,
        scheduler=scheduler,
        device=device
    )
    diffusion.to(device)
    diffusion.train()

    trainer = Trainer(
        diffusion,
        train_loader,
        lr=lr,
        weight_decay=weight_decay,
        steps=steps,
        device=device
    )
    trainer.run_loop()

    # trainer.loss_history.to_csv(os.path.join(parent_dir, 'loss.csv'), index=False)
    # torch.save(diffusion._denoise_fn.state_dict(), os.path.join(parent_dir, 'model.pt'))
    # torch.save(trainer.ema_model.state_dict(), os.path.join(parent_dir, 'model_ema.pt'))

    return model



import numpy as np


def get_custom_dataloader(custom_dataloader: str, dataset, model, scaler, custom_args, verbose: int,
                          x: np.ndarray, y: np.ndarray, train: bool, batch_size: int
                          ):
    from mlc.dataloaders.default import DefaultDataLoader
    from mlc.dataloaders.adversarial.madry import MadryATDataLoader
    DATALOADERS = {"madry": (MadryATDataLoader, {"verbose": verbose}), "subset": (DefaultDataLoader, {"subset": 0.1}),
                   "dist": (DefaultDataLoader, {})}

    dataloader = None
    dataloader_call, args = DATALOADERS.get(custom_dataloader, (None, {}))
    if dataloader_call is None:
        return None

    args = {**args, **custom_args, "dataset": dataset, "model": model, "scaler": scaler}
    dataloader = dataloader_call(**args).get_dataloader(x=x, y=y, train=train, batch_size=batch_size)
    return dataloader



from typing import Any, Dict

from tabsurvey.models import str2model

from mlc.models.model import Model
from mlc.utils import dict2obj


class TabsurveyAdapter(Model):
    def __init__(
        self, tabsurvey_model_name: str, **kwargs: Dict[str, Any]
    ) -> None:

        self.name = tabsurvey_model_name
        self.objective = kwargs["objective"]
        self.constructor_kwargs = kwargs
        args = dict2obj(kwargs)
        params = kwargs

        super().__init__(
            name=self.name, tabsurvey_model_name=tabsurvey_model_name, **kwargs
        )

        self.model = str2model(tabsurvey_model_name)(args=args, params=params)

    def save(self, path: str) -> None:
        pass

    def load(self, path: str) -> None:
        pass


models = [
    ("tabsurvey_adapter", TabsurveyAdapter),
]



from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class NpPdTransformer(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        self.columns = []

    def fit(self, x: pd.DataFrame, y=None):
        self.columns = x.columns
        return self

    def transform(self, x: Union[np.array, pd.DataFrame], y=None):
        if isinstance(x, pd.DataFrame):
            return x

        return pd.DataFrame(x, columns=self.columns)



# -*- coding:utf-8 -*-
import torch
import torch.nn as nn


class Dice(nn.Module):
    """The Data Adaptive Activation Function in DIN,which can be viewed as a generalization of PReLu and can adaptively adjust the rectified point according to distribution of input data.

    Input shape:
        - 2 dims: [batch_size, embedding_size(features)]
        - 3 dims: [batch_size, num_features, embedding_size(features)]

    Output shape:
        - Same shape as input.

    References
        - [Zhou G, Zhu X, Song C, et al. Deep interest network for click-through rate prediction[C]//Proceedings of the 24th ACM SIGKDD International Conference on Knowledge Discovery & Data Mining. ACM, 2018: 1059-1068.](https://arxiv.org/pdf/1706.06978.pdf)
        - https://github.com/zhougr1993/DeepInterestNetwork, https://github.com/fanoping/DIN-pytorch
    """

    def __init__(self, emb_size, dim=2, epsilon=1e-8, device='cpu'):
        super(Dice, self).__init__()
        assert dim == 2 or dim == 3

        self.bn = nn.BatchNorm1d(emb_size, eps=epsilon)
        self.sigmoid = nn.Sigmoid()
        self.dim = dim

        # wrap alpha in nn.Parameter to make it trainable
        if self.dim == 2:
            self.alpha = nn.Parameter(torch.zeros((emb_size,)).to(device))
        else:
            self.alpha = nn.Parameter(torch.zeros((emb_size, 1)).to(device))

    def forward(self, x):
        assert x.dim() == self.dim
        if self.dim == 2:
            x_p = self.sigmoid(self.bn(x))
            out = self.alpha * (1 - x_p) * x + x_p * x
        else:
            x = torch.transpose(x, 1, 2)
            x_p = self.sigmoid(self.bn(x))
            out = self.alpha * (1 - x_p) * x + x_p * x
            out = torch.transpose(out, 1, 2)
        return out


class Identity(nn.Module):

    def __init__(self, **kwargs):
        super(Identity, self).__init__()

    def forward(self, inputs):
        return inputs


def activation_layer(act_name, hidden_size=None, dice_dim=2):
    """Construct activation layers

    Args:
        act_name: str or nn.Module, name of activation function
        hidden_size: int, used for Dice activation
        dice_dim: int, used for Dice activation
    Return:
        act_layer: activation layer
    """
    if isinstance(act_name, str):
        if act_name.lower() == 'sigmoid':
            act_layer = nn.Sigmoid()
        elif act_name.lower() == 'linear':
            act_layer = Identity()
        elif act_name.lower() == 'relu':
            act_layer = nn.ReLU(inplace=True)
        elif act_name.lower() == 'dice':
            assert dice_dim
            act_layer = Dice(hidden_size, dice_dim)
        elif act_name.lower() == 'prelu':
            act_layer = nn.PReLU()
    elif issubclass(act_name, nn.Module):
        act_layer = act_name()
    else:
        raise NotImplementedError

    return act_layer


if __name__ == "__main__":
    pass



import pandas as pd

from mlc.datasets.dataset import (
    CsvDataSource,
    Dataset,
    DateSorter,
    DefaultSplitter,
    Task,
    TaskProcessor,
)


class RegressionClassificationProcessor(TaskProcessor):
    def transform(self, task_data: pd.Series) -> pd.Series:
        return (task_data > 0).astype(int)


def create_dataset(is_classification: bool) -> Dataset:
    data_source = CsvDataSource(
        path="./data/mlc/airlines/flight_delay_ord.csv"
    )
    metadata_source = CsvDataSource(
        path="./data/mlc/airlines/flight_delay_ord_metadata.csv"
    )
    sorter = DateSorter(date_col="Date")
    splitter = DefaultSplitter()

    if is_classification:
        name = "flight_delay_ord_classification"
        task = Task(
            name="ArrDelay",
            task_type="classification",
            evaluation_metric="f1_score",
        )
    else:
        name = "flight_delay_ord_regression"
        task = Task(
            name="ArrDelay",
            task_type="regression",
            evaluation_metric="rmse",
            task_processor=RegressionClassificationProcessor(),
        )

    dataset = Dataset(
        name=name,
        data_source=data_source,
        metadata_source=metadata_source,
        tasks=[task],
        sorter=sorter,
        splitter=splitter,
        relation_constraints=[],
    )

    return dataset


datasets = [
    {
        "name": "flight_delay_ord_classification",
        "fun_create": lambda x: create_dataset(is_classification=True),
    },
    {
        "name": "flight_delay_ord_regression",
        "fun_create": lambda x: create_dataset(is_classification=False),
    },
]



from sdv.metadata import SingleTableMetadata
from sdv.single_table import CopulaGANSynthesizer
from sdv.single_table import GaussianCopulaSynthesizer

from mlc.dataloaders.generative.ctgan import CTGANDataLoader
from mlc.dataloaders.sdv import SDVDataLoader
from mlc.datasets.dataset import Dataset


class GaussianCopulaDataLoader(CTGANDataLoader):
    def __init__(self, num_workers: int = 4, original_dataset: Dataset = None, scaler=None, copula_params: dict = {},
                 filter_constraints: bool = False, pretrained_path: str = None, save_path: str = None) -> None:
        self.num_workers = num_workers
        self.dataset = original_dataset
        self.scaler = scaler
        self.filter_constraints = filter_constraints

        metadata = SingleTableMetadata()
        metadata.detect_from_dataframe(data=self.dataset.get_data())

        self.generator = GaussianCopulaSynthesizer(metadata, **copula_params)
        SDVDataLoader.__init__(num_workers, pretrained_path, save_path)


class GANCopulaDataLoader(CTGANDataLoader):
    def __init__(self, num_workers: int = 4, original_dataset: Dataset = None, scaler=None, copula_params: dict = {},
                 filter_constraints: bool = False, pretrained_path: str = None, save_path: str = None) -> None:
        self.num_workers = num_workers
        self.dataset = original_dataset
        self.scaler = scaler
        self.filter_constraints = filter_constraints

        metadata = SingleTableMetadata()
        metadata.detect_from_dataframe(data=self.dataset.get_data())

        self.generator = CopulaGANSynthesizer(metadata, **copula_params)
        SDVDataLoader.__init__(num_workers, pretrained_path, save_path)



from typing import List, Optional, Union

import numpy as np

from mlc.constraints.backend import Backend
from mlc.typing import NDNumber


class NumpyBackend(Backend):
    def __init__(self, eps: float = 0.000001) -> None:
        self.eps = np.array(eps)

    def get_eps(self) -> NDNumber:
        return self.eps

    def get_zeros(self, operands: List[NDNumber]) -> NDNumber:
        i = np.argmax([op.shape[-1] for op in operands])
        return np.zeros(operands[i].shape, dtype=operands[i].dtype)

    # Values
    def constant(self, value: Union[int, float]) -> NDNumber:
        return np.array([value])

    def feature(self, x: NDNumber, feature_id: int) -> NDNumber:
        return x[:, feature_id]

    # Math operations

    def math_operation(
        self,
        operator: str,
        left_operand: NDNumber,
        right_operand: NDNumber,
    ) -> NDNumber:
        if operator == "+":
            return left_operand + right_operand
        elif operator == "-":
            return left_operand - right_operand
        elif operator == "*":
            return left_operand * right_operand
        elif operator == "/":
            return left_operand / right_operand
        elif operator == "**":
            return left_operand**right_operand
        elif operator == "%":
            return left_operand % right_operand
        else:
            raise ValueError(f"Unknown operator: {operator}")

    def safe_division(
        self,
        dividend: NDNumber,
        divisor: NDNumber,
        safe_value: NDNumber,
    ) -> NDNumber:
        return np.divide(
            dividend,
            divisor,
            out=np.full_like(dividend, safe_value),
            where=divisor != 0,
        )

    def log(
        self, operand: NDNumber, safe_value: Optional[NDNumber] = None
    ) -> NDNumber:
        if safe_value is not None:
            return np.log(
                operand,
                out=np.full_like(operand, fill_value=safe_value),
                where=(operand > 0),
            )
        return np.log(operand)

    def many_sum(self, operands: List[NDNumber]) -> NDNumber:
        return np.sum(operands, axis=0)

    def less_equal_constraint(
        self, left_operand: NDNumber, right_operand: NDNumber
    ) -> NDNumber:

        zeros = self.get_zeros([left_operand, right_operand])
        substraction = left_operand - right_operand
        bound_zero = np.max(np.stack([zeros, substraction]), axis=0)

        return bound_zero

    def less_constraint(
        self, left_operand: NDNumber, right_operand: NDNumber
    ) -> NDNumber:
        zeros = self.get_zeros([left_operand, right_operand])
        substraction = (left_operand + self.eps) - right_operand
        bound_zero = np.max(np.stack([zeros, substraction]), axis=0)
        return bound_zero

    def equal_constraint(
        self, left_operand: NDNumber, right_operand: NDNumber
    ) -> NDNumber:
        return np.abs(left_operand - right_operand)

    def or_constraint(self, operands: List[NDNumber]) -> NDNumber:
        return np.min(np.stack(operands), axis=0)

    def and_constraint(self, operands: List[NDNumber]) -> NDNumber:
        return np.sum(np.stack(operands), axis=0)

    def count(self, operands: List[NDNumber], inverse: bool) -> NDNumber:
        if inverse:
            return np.sum(
                np.stack([(op != 0).astype(float) for op in operands]),
                axis=0,
            )
        else:
            return np.sum(
                np.stack([(op == 0).astype(float) for op in operands]),
                axis=0,
            )



from typing import Optional, Union

import numpy as np
import numpy.typing as npt


def get_feature_index(
    feature_names: Optional[npt.ArrayLike], feature_id: Union[int, str]
) -> int:
    if isinstance(feature_id, str):
        if feature_names is None:
            raise ValueError(
                f"Feature names not provided. "
                f"Impossible to convert {feature_id} to index"
            )

        feature_names = np.array(feature_names)
        index = np.where(feature_names == feature_id)[0]

        if len(index) <= 0:
            raise IndexError(f"{feature_id} is not in {feature_names}")

        return index[0]

    else:
        return feature_id



import itertools

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..layers.activation import activation_layer
from ..layers.core import Conv2dSame
from ..layers.sequence import KMaxPooling


class FM(nn.Module):
    """Factorization Machine models pairwise (order-2) feature interactions
     without linear term and bias.
      Input shape
        - 3D tensor with shape: ``(batch_size,field_size,embedding_size)``.
      Output shape
        - 2D tensor with shape: ``(batch_size, 1)``.
      References
        - [Factorization Machines](https://www.csie.ntu.edu.tw/~b97053/paper/Rendle2010FM.pdf)
    """

    def __init__(self):
        super(FM, self).__init__()

    def forward(self, inputs):
        fm_input = inputs

        square_of_sum = torch.pow(torch.sum(fm_input, dim=1, keepdim=True), 2)
        sum_of_square = torch.sum(fm_input * fm_input, dim=1, keepdim=True)
        cross_term = square_of_sum - sum_of_square
        cross_term = 0.5 * torch.sum(cross_term, dim=2, keepdim=False)

        return cross_term


class BiInteractionPooling(nn.Module):
    """Bi-Interaction Layer used in Neural FM,compress the
     pairwise element-wise product of features into one single vector.

      Input shape
        - A 3D tensor with shape:``(batch_size,field_size,embedding_size)``.

      Output shape
        - 3D tensor with shape: ``(batch_size,1,embedding_size)``.

      References
        - [He X, Chua T S. Neural factorization machines for sparse predictive analytics[C]//Proceedings of the 40th International ACM SIGIR conference on Research and Development in Information Retrieval. ACM, 2017: 355-364.](http://arxiv.org/abs/1708.05027)
    """

    def __init__(self):
        super(BiInteractionPooling, self).__init__()

    def forward(self, inputs):
        concated_embeds_value = inputs
        square_of_sum = torch.pow(
            torch.sum(concated_embeds_value, dim=1, keepdim=True), 2)
        sum_of_square = torch.sum(
            concated_embeds_value * concated_embeds_value, dim=1, keepdim=True)
        cross_term = 0.5 * (square_of_sum - sum_of_square)
        return cross_term


class SENETLayer(nn.Module):
    """SENETLayer used in FiBiNET.
      Input shape
        - A list of 3D tensor with shape: ``(batch_size,filed_size,embedding_size)``.
      Output shape
        - A list of 3D tensor with shape: ``(batch_size,filed_size,embedding_size)``.
      Arguments
        - **filed_size** : Positive integer, number of feature groups.
        - **reduction_ratio** : Positive integer, dimensionality of the
         attention network output space.
        - **seed** : A Python integer to use as random seed.
      References
        - [FiBiNET: Combining Feature Importance and Bilinear feature Interaction for Click-Through Rate Prediction
Tongwen](https://arxiv.org/pdf/1905.09433.pdf)
    """

    def __init__(self, filed_size, reduction_ratio=3, seed=1024, device='cpu'):
        super(SENETLayer, self).__init__()
        self.seed = seed
        self.filed_size = filed_size
        self.reduction_size = max(1, filed_size // reduction_ratio)
        self.excitation = nn.Sequential(
            nn.Linear(self.filed_size, self.reduction_size, bias=False),
            nn.ReLU(),
            nn.Linear(self.reduction_size, self.filed_size, bias=False),
            nn.ReLU()
        )
        self.to(device)

    def forward(self, inputs):
        if len(inputs.shape) != 3:
            raise ValueError(
                "Unexpected inputs dimensions %d, expect to be 3 dimensions" % (len(inputs.shape)))
        Z = torch.mean(inputs, dim=-1, out=None)
        A = self.excitation(Z)
        V = torch.mul(inputs, torch.unsqueeze(A, dim=2))

        return V


class BilinearInteraction(nn.Module):
    """BilinearInteraction Layer used in FiBiNET.
      Input shape
        - A list of 3D tensor with shape: ``(batch_size,filed_size, embedding_size)``.
      Output shape
        - 3D tensor with shape: ``(batch_size,filed_size*(filed_size-1)/2, embedding_size)``.
      Arguments
        - **filed_size** : Positive integer, number of feature groups.
        - **embedding_size** : Positive integer, embedding size of sparse features.
        - **bilinear_type** : String, types of bilinear functions used in this layer.
        - **seed** : A Python integer to use as random seed.
      References
        - [FiBiNET: Combining Feature Importance and Bilinear feature Interaction for Click-Through Rate Prediction
Tongwen](https://arxiv.org/pdf/1905.09433.pdf)
    """

    def __init__(self, filed_size, embedding_size, bilinear_type="interaction", seed=1024, device='cpu'):
        super(BilinearInteraction, self).__init__()
        self.bilinear_type = bilinear_type
        self.seed = seed
        self.bilinear = nn.ModuleList()
        if self.bilinear_type == "all":
            self.bilinear = nn.Linear(
                embedding_size, embedding_size, bias=False)
        elif self.bilinear_type == "each":
            for _ in range(filed_size):
                self.bilinear.append(
                    nn.Linear(embedding_size, embedding_size, bias=False))
        elif self.bilinear_type == "interaction":
            for _, _ in itertools.combinations(range(filed_size), 2):
                self.bilinear.append(
                    nn.Linear(embedding_size, embedding_size, bias=False))
        else:
            raise NotImplementedError
        self.to(device)

    def forward(self, inputs):
        if len(inputs.shape) != 3:
            raise ValueError(
                "Unexpected inputs dimensions %d, expect to be 3 dimensions" % (len(inputs.shape)))
        inputs = torch.split(inputs, 1, dim=1)
        if self.bilinear_type == "all":
            p = [torch.mul(self.bilinear(v_i), v_j)
                 for v_i, v_j in itertools.combinations(inputs, 2)]
        elif self.bilinear_type == "each":
            p = [torch.mul(self.bilinear[i](inputs[i]), inputs[j])
                 for i, j in itertools.combinations(range(len(inputs)), 2)]
        elif self.bilinear_type == "interaction":
            p = [torch.mul(bilinear(v[0]), v[1])
                 for v, bilinear in zip(itertools.combinations(inputs, 2), self.bilinear)]
        else:
            raise NotImplementedError
        return torch.cat(p, dim=1)


class CIN(nn.Module):
    """Compressed Interaction Network used in xDeepFM.
      Input shape
        - 3D tensor with shape: ``(batch_size,field_size,embedding_size)``.
      Output shape
        - 2D tensor with shape: ``(batch_size, featuremap_num)`` ``featuremap_num =  sum(self.layer_size[:-1]) // 2 + self.layer_size[-1]`` if ``split_half=True``,else  ``sum(layer_size)`` .
      Arguments
        - **filed_size** : Positive integer, number of feature groups.
        - **layer_size** : list of int.Feature maps in each layer.
        - **activation** : activation function name used on feature maps.
        - **split_half** : bool.if set to False, half of the feature maps in each hidden will connect to output unit.
        - **seed** : A Python integer to use as random seed.
      References
        - [Lian J, Zhou X, Zhang F, et al. xDeepFM: Combining Explicit and Implicit Feature Interactions for Recommender Systems[J]. arXiv preprint arXiv:1803.05170, 2018.] (https://arxiv.org/pdf/1803.05170.pdf)
    """

    def __init__(self, field_size, layer_size=(128, 128), activation='relu', split_half=True, l2_reg=1e-5, seed=1024,
                 device='cpu'):
        super(CIN, self).__init__()
        if len(layer_size) == 0:
            raise ValueError(
                "layer_size must be a list(tuple) of length greater than 1")

        self.layer_size = layer_size
        self.field_nums = [field_size]
        self.split_half = split_half
        self.activation = activation_layer(activation)
        self.l2_reg = l2_reg
        self.seed = seed

        self.conv1ds = nn.ModuleList()
        for i, size in enumerate(self.layer_size):
            self.conv1ds.append(
                nn.Conv1d(self.field_nums[-1] * self.field_nums[0], size, 1))

            if self.split_half:
                if i != len(self.layer_size) - 1 and size % 2 > 0:
                    raise ValueError(
                        "layer_size must be even number except for the last layer when split_half=True")

                self.field_nums.append(size // 2)
            else:
                self.field_nums.append(size)

        #         for tensor in self.conv1ds:
        #             nn.init.normal_(tensor.weight, mean=0, std=init_std)
        self.to(device)

    def forward(self, inputs):
        if len(inputs.shape) != 3:
            raise ValueError(
                "Unexpected inputs dimensions %d, expect to be 3 dimensions" % (len(inputs.shape)))
        batch_size = inputs.shape[0]
        dim = inputs.shape[-1]
        hidden_nn_layers = [inputs]
        final_result = []

        for i, size in enumerate(self.layer_size):
            # x^(k-1) * x^0
            x = torch.einsum(
                'bhd,bmd->bhmd', hidden_nn_layers[-1], hidden_nn_layers[0])
            # x.shape = (batch_size , hi * m, dim)
            x = x.reshape(
                batch_size, hidden_nn_layers[-1].shape[1] * hidden_nn_layers[0].shape[1], dim)
            # x.shape = (batch_size , hi, dim)
            x = self.conv1ds[i](x)

            if self.activation is None or self.activation == 'linear':
                curr_out = x
            else:
                curr_out = self.activation(x)

            if self.split_half:
                if i != len(self.layer_size) - 1:
                    next_hidden, direct_connect = torch.split(
                        curr_out, 2 * [size // 2], 1)
                else:
                    direct_connect = curr_out
                    next_hidden = 0
            else:
                direct_connect = curr_out
                next_hidden = curr_out

            final_result.append(direct_connect)
            hidden_nn_layers.append(next_hidden)

        result = torch.cat(final_result, dim=1)
        result = torch.sum(result, -1)

        return result


class AFMLayer(nn.Module):
    """Attentonal Factorization Machine models pairwise (order-2) feature
    interactions without linear term and bias.
      Input shape
        - A list of 3D tensor with shape: ``(batch_size,1,embedding_size)``.
      Output shape
        - 2D tensor with shape: ``(batch_size, 1)``.
      Arguments
        - **in_features** : Positive integer, dimensionality of input features.
        - **attention_factor** : Positive integer, dimensionality of the
         attention network output space.
        - **l2_reg_w** : float between 0 and 1. L2 regularizer strength
         applied to attention network.
        - **dropout_rate** : float between in [0,1). Fraction of the attention net output units to dropout.
        - **seed** : A Python integer to use as random seed.
      References
        - [Attentional Factorization Machines : Learning the Weight of Feature
        Interactions via Attention Networks](https://arxiv.org/pdf/1708.04617.pdf)
    """

    def __init__(self, in_features, attention_factor=4, l2_reg_w=0, dropout_rate=0, seed=1024, device='cpu'):
        super(AFMLayer, self).__init__()
        self.attention_factor = attention_factor
        self.l2_reg_w = l2_reg_w
        self.dropout_rate = dropout_rate
        self.seed = seed
        embedding_size = in_features

        self.attention_W = nn.Parameter(torch.Tensor(
            embedding_size, self.attention_factor))

        self.attention_b = nn.Parameter(torch.Tensor(self.attention_factor))

        self.projection_h = nn.Parameter(
            torch.Tensor(self.attention_factor, 1))

        self.projection_p = nn.Parameter(torch.Tensor(embedding_size, 1))

        for tensor in [self.attention_W, self.projection_h, self.projection_p]:
            nn.init.xavier_normal_(tensor, )

        for tensor in [self.attention_b]:
            nn.init.zeros_(tensor, )

        self.dropout = nn.Dropout(dropout_rate)

        self.to(device)

    def forward(self, inputs):
        embeds_vec_list = inputs
        row = []
        col = []

        for r, c in itertools.combinations(embeds_vec_list, 2):
            row.append(r)
            col.append(c)

        p = torch.cat(row, dim=1)
        q = torch.cat(col, dim=1)
        inner_product = p * q

        bi_interaction = inner_product
        attention_temp = F.relu(torch.tensordot(
            bi_interaction, self.attention_W, dims=([-1], [0])) + self.attention_b)

        self.normalized_att_score = F.softmax(torch.tensordot(
            attention_temp, self.projection_h, dims=([-1], [0])), dim=1)
        attention_output = torch.sum(
            self.normalized_att_score * bi_interaction, dim=1)

        attention_output = self.dropout(attention_output)  # training

        afm_out = torch.tensordot(
            attention_output, self.projection_p, dims=([-1], [0]))
        return afm_out


class InteractingLayer(nn.Module):
    """A Layer used in AutoInt that model the correlations between different feature fields by multi-head self-attention mechanism.
      Input shape
            - A 3D tensor with shape: ``(batch_size,field_size,embedding_size)``.
      Output shape
            - 3D tensor with shape:``(batch_size,field_size,embedding_size)``.
      Arguments
            - **in_features** : Positive integer, dimensionality of input features.
            - **head_num**: int.The head number in multi-head self-attention network.
            - **use_res**: bool.Whether or not use standard residual connections before output.
            - **seed**: A Python integer to use as random seed.
      References
            - [Song W, Shi C, Xiao Z, et al. AutoInt: Automatic Feature Interaction Learning via Self-Attentive Neural Networks[J]. arXiv preprint arXiv:1810.11921, 2018.](https://arxiv.org/abs/1810.11921)
    """

    def __init__(self, embedding_size, head_num=2, use_res=True, scaling=False, seed=1024, device='cpu'):
        super(InteractingLayer, self).__init__()
        if head_num <= 0:
            raise ValueError('head_num must be a int > 0')
        if embedding_size % head_num != 0:
            raise ValueError('embedding_size is not an integer multiple of head_num!')
        self.att_embedding_size = embedding_size // head_num
        self.head_num = head_num
        self.use_res = use_res
        self.scaling = scaling
        self.seed = seed

        self.W_Query = nn.Parameter(torch.Tensor(embedding_size, embedding_size))
        self.W_key = nn.Parameter(torch.Tensor(embedding_size, embedding_size))
        self.W_Value = nn.Parameter(torch.Tensor(embedding_size, embedding_size))

        if self.use_res:
            self.W_Res = nn.Parameter(torch.Tensor(embedding_size, embedding_size))
        for tensor in self.parameters():
            nn.init.normal_(tensor, mean=0.0, std=0.05)

        self.to(device)

    def forward(self, inputs):

        if len(inputs.shape) != 3:
            raise ValueError(
                "Unexpected inputs dimensions %d, expect to be 3 dimensions" % (len(inputs.shape)))

        # None F D
        querys = torch.tensordot(inputs, self.W_Query, dims=([-1], [0]))
        keys = torch.tensordot(inputs, self.W_key, dims=([-1], [0]))
        values = torch.tensordot(inputs, self.W_Value, dims=([-1], [0]))

        # head_num None F D/head_num
        querys = torch.stack(torch.split(querys, self.att_embedding_size, dim=2))
        keys = torch.stack(torch.split(keys, self.att_embedding_size, dim=2))
        values = torch.stack(torch.split(values, self.att_embedding_size, dim=2))

        inner_product = torch.einsum('bnik,bnjk->bnij', querys, keys)  # head_num None F F
        if self.scaling:
            inner_product /= self.att_embedding_size ** 0.5
        self.normalized_att_scores = F.softmax(inner_product, dim=-1)  # head_num None F F
        result = torch.matmul(self.normalized_att_scores, values)  # head_num None F D/head_num

        result = torch.cat(torch.split(result, 1, ), dim=-1)
        result = torch.squeeze(result, dim=0)  # None F D
        if self.use_res:
            result += torch.tensordot(inputs, self.W_Res, dims=([-1], [0]))
        result = F.relu(result)

        return result


class CrossNet(nn.Module):
    """The Cross Network part of Deep&Cross Network model,
    which leans both low and high degree cross feature.
      Input shape
        - 2D tensor with shape: ``(batch_size, units)``.
      Output shape
        - 2D tensor with shape: ``(batch_size, units)``.
      Arguments
        - **in_features** : Positive integer, dimensionality of input features.
        - **input_feature_num**: Positive integer, shape(Input tensor)[-1]
        - **layer_num**: Positive integer, the cross layer number
        - **parameterization**: string, ``"vector"``  or ``"matrix"`` ,  way to parameterize the cross network.
        - **l2_reg**: float between 0 and 1. L2 regularizer strength applied to the kernel weights matrix
        - **seed**: A Python integer to use as random seed.
      References
        - [Wang R, Fu B, Fu G, et al. Deep & cross network for ad click predictions[C]//Proceedings of the ADKDD'17. ACM, 2017: 12.](https://arxiv.org/abs/1708.05123)
        - [Wang R, Shivanna R, Cheng D Z, et al. DCN-M: Improved Deep & Cross Network for Feature Cross Learning in Web-scale Learning to Rank Systems[J]. 2020.](https://arxiv.org/abs/2008.13535)
    """

    def __init__(self, in_features, layer_num=2, parameterization='vector', seed=1024, device='cpu'):
        super(CrossNet, self).__init__()
        self.layer_num = layer_num
        self.parameterization = parameterization
        if self.parameterization == 'vector':
            # weight in DCN.  (in_features, 1)
            self.kernels = nn.Parameter(torch.Tensor(self.layer_num, in_features, 1))
        elif self.parameterization == 'matrix':
            # weight matrix in DCN-M.  (in_features, in_features)
            self.kernels = nn.Parameter(torch.Tensor(self.layer_num, in_features, in_features))
        else:  # error
            raise ValueError("parameterization should be 'vector' or 'matrix'")

        self.bias = nn.Parameter(torch.Tensor(self.layer_num, in_features, 1))

        for i in range(self.kernels.shape[0]):
            nn.init.xavier_normal_(self.kernels[i])
        for i in range(self.bias.shape[0]):
            nn.init.zeros_(self.bias[i])

        self.to(device)

    def forward(self, inputs):
        x_0 = inputs.unsqueeze(2)
        x_l = x_0
        for i in range(self.layer_num):
            if self.parameterization == 'vector':
                xl_w = torch.tensordot(x_l, self.kernels[i], dims=([1], [0]))
                dot_ = torch.matmul(x_0, xl_w)
                x_l = dot_ + self.bias[i] + x_l
            elif self.parameterization == 'matrix':
                xl_w = torch.matmul(self.kernels[i], x_l)  # W * xi  (bs, in_features, 1)
                dot_ = xl_w + self.bias[i]  # W * xi + b
                x_l = x_0 * dot_ + x_l  # x0 · (W * xi + b) +xl  Hadamard-product
            else:  # error
                raise ValueError("parameterization should be 'vector' or 'matrix'")
        x_l = torch.squeeze(x_l, dim=2)
        return x_l


class CrossNetMix(nn.Module):
    """The Cross Network part of DCN-Mix model, which improves DCN-M by:
      1 add MOE to learn feature interactions in different subspaces
      2 add nonlinear transformations in low-dimensional space
      Input shape
        - 2D tensor with shape: ``(batch_size, units)``.
      Output shape
        - 2D tensor with shape: ``(batch_size, units)``.
      Arguments
        - **in_features** : Positive integer, dimensionality of input features.
        - **low_rank** : Positive integer, dimensionality of low-rank sapce.
        - **num_experts** : Positive integer, number of experts.
        - **layer_num**: Positive integer, the cross layer number
        - **device**: str, e.g. ``"cpu"`` or ``"cuda:0"``
      References
        - [Wang R, Shivanna R, Cheng D Z, et al. DCN-M: Improved Deep & Cross Network for Feature Cross Learning in Web-scale Learning to Rank Systems[J]. 2020.](https://arxiv.org/abs/2008.13535)
    """

    def __init__(self, in_features, low_rank=32, num_experts=4, layer_num=2, device='cpu'):
        super(CrossNetMix, self).__init__()
        self.layer_num = layer_num
        self.num_experts = num_experts

        # U: (in_features, low_rank)
        self.U_list = nn.Parameter(torch.Tensor(self.layer_num, num_experts, in_features, low_rank))
        # V: (in_features, low_rank)
        self.V_list = nn.Parameter(torch.Tensor(self.layer_num, num_experts, in_features, low_rank))
        # C: (low_rank, low_rank)
        self.C_list = nn.Parameter(torch.Tensor(self.layer_num, num_experts, low_rank, low_rank))
        self.gating = nn.ModuleList([nn.Linear(in_features, 1, bias=False) for i in range(self.num_experts)])

        self.bias = nn.Parameter(torch.Tensor(self.layer_num, in_features, 1))

        init_para_list = [self.U_list, self.V_list, self.C_list]
        for para in init_para_list:
            for i in range(self.layer_num):
                nn.init.xavier_normal_(para[i])

        for i in range(len(self.bias)):
            nn.init.zeros_(self.bias[i])

        self.to(device)

    def forward(self, inputs):
        x_0 = inputs.unsqueeze(2)  # (bs, in_features, 1)
        x_l = x_0
        for i in range(self.layer_num):
            output_of_experts = []
            gating_score_of_experts = []
            for expert_id in range(self.num_experts):
                # (1) G(x_l)
                # compute the gating score by x_l
                gating_score_of_experts.append(self.gating[expert_id](x_l.squeeze(2)))

                # (2) E(x_l)
                # project the input x_l to $\mathbb{R}^{r}$
                v_x = torch.matmul(self.V_list[i][expert_id].t(), x_l)  # (bs, low_rank, 1)

                # nonlinear activation in low rank space
                v_x = torch.tanh(v_x)
                v_x = torch.matmul(self.C_list[i][expert_id], v_x)
                v_x = torch.tanh(v_x)

                # project back to $\mathbb{R}^{d}$
                uv_x = torch.matmul(self.U_list[i][expert_id], v_x)  # (bs, in_features, 1)

                dot_ = uv_x + self.bias[i]
                dot_ = x_0 * dot_  # Hadamard-product

                output_of_experts.append(dot_.squeeze(2))

            # (3) mixture of low-rank experts
            output_of_experts = torch.stack(output_of_experts, 2)  # (bs, in_features, num_experts)
            gating_score_of_experts = torch.stack(gating_score_of_experts, 1)  # (bs, num_experts, 1)
            moe_out = torch.matmul(output_of_experts, gating_score_of_experts.softmax(1))
            x_l = moe_out + x_l  # (bs, in_features, 1)

        x_l = x_l.squeeze()  # (bs, in_features)
        return x_l


class InnerProductLayer(nn.Module):
    """InnerProduct Layer used in PNN that compute the element-wise
    product or inner product between feature vectors.
      Input shape
        - a list of 3D tensor with shape: ``(batch_size,1,embedding_size)``.
      Output shape
        - 3D tensor with shape: ``(batch_size, N*(N-1)/2 ,1)`` if use reduce_sum. or 3D tensor with shape:
        ``(batch_size, N*(N-1)/2, embedding_size )`` if not use reduce_sum.
      Arguments
        - **reduce_sum**: bool. Whether return inner product or element-wise product
      References
            - [Qu Y, Cai H, Ren K, et al. Product-based neural networks for user response prediction[C]//
            Data Mining (ICDM), 2016 IEEE 16th International Conference on. IEEE, 2016: 1149-1154.]
            (https://arxiv.org/pdf/1611.00144.pdf)"""

    def __init__(self, reduce_sum=True, device='cpu'):
        super(InnerProductLayer, self).__init__()
        self.reduce_sum = reduce_sum
        self.to(device)

    def forward(self, inputs):

        embed_list = inputs
        row = []
        col = []
        num_inputs = len(embed_list)

        for i in range(num_inputs - 1):
            for j in range(i + 1, num_inputs):
                row.append(i)
                col.append(j)
        p = torch.cat([embed_list[idx]
                       for idx in row], dim=1)  # batch num_pairs k
        q = torch.cat([embed_list[idx]
                       for idx in col], dim=1)

        inner_product = p * q
        if self.reduce_sum:
            inner_product = torch.sum(
                inner_product, dim=2, keepdim=True)
        return inner_product


class OutterProductLayer(nn.Module):
    """OutterProduct Layer used in PNN.This implemention is
    adapted from code that the author of the paper published on https://github.com/Atomu2014/product-nets.
      Input shape
            - A list of N 3D tensor with shape: ``(batch_size,1,embedding_size)``.
      Output shape
            - 2D tensor with shape:``(batch_size,N*(N-1)/2 )``.
      Arguments
            - **filed_size** : Positive integer, number of feature groups.
            - **kernel_type**: str. The kernel weight matrix type to use,can be mat,vec or num
            - **seed**: A Python integer to use as random seed.
      References
            - [Qu Y, Cai H, Ren K, et al. Product-based neural networks for user response prediction[C]//Data Mining (ICDM), 2016 IEEE 16th International Conference on. IEEE, 2016: 1149-1154.](https://arxiv.org/pdf/1611.00144.pdf)
    """

    def __init__(self, field_size, embedding_size, kernel_type='mat', seed=1024, device='cpu'):
        super(OutterProductLayer, self).__init__()
        self.kernel_type = kernel_type

        num_inputs = field_size
        num_pairs = int(num_inputs * (num_inputs - 1) / 2)
        embed_size = embedding_size
        if self.kernel_type == 'mat':

            self.kernel = nn.Parameter(torch.Tensor(
                embed_size, num_pairs, embed_size))

        elif self.kernel_type == 'vec':
            self.kernel = nn.Parameter(torch.Tensor(num_pairs, embed_size))

        elif self.kernel_type == 'num':
            self.kernel = nn.Parameter(torch.Tensor(num_pairs, 1))
        nn.init.xavier_uniform_(self.kernel)

        self.to(device)

    def forward(self, inputs):
        embed_list = inputs
        row = []
        col = []
        num_inputs = len(embed_list)
        for i in range(num_inputs - 1):
            for j in range(i + 1, num_inputs):
                row.append(i)
                col.append(j)
        p = torch.cat([embed_list[idx]
                       for idx in row], dim=1)  # batch num_pairs k
        q = torch.cat([embed_list[idx] for idx in col], dim=1)

        # -------------------------
        if self.kernel_type == 'mat':
            p.unsqueeze_(dim=1)
            # k     k* pair* k
            # batch * pair
            kp = torch.sum(

                # batch * pair * k

                torch.mul(

                    # batch * pair * k

                    torch.transpose(

                        # batch * k * pair

                        torch.sum(

                            # batch * k * pair * k

                            torch.mul(

                                p, self.kernel),

                            dim=-1),

                        2, 1),

                    q),

                dim=-1)
        else:
            # 1 * pair * (k or 1)

            k = torch.unsqueeze(self.kernel, 0)

            # batch * pair

            kp = torch.sum(p * q * k, dim=-1)

            # p q # b * p * k

        return kp


class ConvLayer(nn.Module):
    """Conv Layer used in CCPM.

      Input shape
            - A list of N 3D tensor with shape: ``(batch_size,1,filed_size,embedding_size)``.
      Output shape
            - A list of N 3D tensor with shape: ``(batch_size,last_filters,pooling_size,embedding_size)``.
      Arguments
            - **filed_size** : Positive integer, number of feature groups.
            - **conv_kernel_width**: list. list of positive integer or empty list,the width of filter in each conv layer.
            - **conv_filters**: list. list of positive integer or empty list,the number of filters in each conv layer.
      Reference:
            - Liu Q, Yu F, Wu S, et al. A convolutional click prediction model[C]//Proceedings of the 24th ACM International on Conference on Information and Knowledge Management. ACM, 2015: 1743-1746.(http://ir.ia.ac.cn/bitstream/173211/12337/1/A%20Convolutional%20Click%20Prediction%20Model.pdf)
    """

    def __init__(self, field_size, conv_kernel_width, conv_filters, device='cpu'):
        super(ConvLayer, self).__init__()
        self.device = device
        module_list = []
        n = int(field_size)
        l = len(conv_filters)
        filed_shape = n
        for i in range(1, l + 1):
            if i == 1:
                in_channels = 1
            else:
                in_channels = conv_filters[i - 2]
            out_channels = conv_filters[i - 1]
            width = conv_kernel_width[i - 1]
            k = max(1, int((1 - pow(i / l, l - i)) * n)) if i < l else 3
            module_list.append(Conv2dSame(in_channels=in_channels, out_channels=out_channels, kernel_size=(width, 1),
                                          stride=1).to(self.device))
            module_list.append(torch.nn.Tanh().to(self.device))

            # KMaxPooling, extract top_k, returns tensors values
            module_list.append(KMaxPooling(k=min(k, filed_shape), axis=2, device=self.device).to(self.device))
            filed_shape = min(k, filed_shape)
        self.conv_layer = nn.Sequential(*module_list)
        self.to(device)
        self.filed_shape = filed_shape

    def forward(self, inputs):
        return self.conv_layer(inputs)


class LogTransformLayer(nn.Module):
    """Logarithmic Transformation Layer in Adaptive factorization network, which models arbitrary-order cross features.

      Input shape
        - 3D tensor with shape: ``(batch_size, field_size, embedding_size)``.
      Output shape
        - 2D tensor with shape: ``(batch_size, ltl_hidden_size*embedding_size)``.
      Arguments
        - **field_size** : positive integer, number of feature groups
        - **embedding_size** : positive integer, embedding size of sparse features
        - **ltl_hidden_size** : integer, the number of logarithmic neurons in AFN
      References
        - Cheng, W., Shen, Y. and Huang, L. 2020. Adaptive Factorization Network: Learning Adaptive-Order Feature
         Interactions. Proceedings of the AAAI Conference on Artificial Intelligence. 34, 04 (Apr. 2020), 3609-3616.
    """

    def __init__(self, field_size, embedding_size, ltl_hidden_size):
        super(LogTransformLayer, self).__init__()

        self.ltl_weights = nn.Parameter(torch.Tensor(field_size, ltl_hidden_size))
        self.ltl_biases = nn.Parameter(torch.Tensor(1, 1, ltl_hidden_size))
        self.bn = nn.ModuleList([nn.BatchNorm1d(embedding_size) for i in range(2)])
        nn.init.normal_(self.ltl_weights, mean=0.0, std=0.1)
        nn.init.zeros_(self.ltl_biases, )

    def forward(self, inputs):
        # Avoid numeric overflow
        afn_input = torch.clamp(torch.abs(inputs), min=1e-7, max=float("Inf"))
        # Transpose to shape: ``(batch_size,embedding_size,field_size)``
        afn_input_trans = torch.transpose(afn_input, 1, 2)
        # Logarithmic transformation layer
        ltl_result = torch.log(afn_input_trans)
        ltl_result = self.bn[0](ltl_result)
        ltl_result = torch.matmul(ltl_result, self.ltl_weights) + self.ltl_biases
        ltl_result = torch.exp(ltl_result)
        ltl_result = self.bn[1](ltl_result)
        ltl_result = torch.flatten(ltl_result, start_dim=1)
        return ltl_result



from mlc.models.model import Model
from mlc.models.samples.sklearn_default import RandomForest as DefaultRf

rf_parameters = {
    "n_estimators": 125,
    "min_samples_split": 6,
    "min_samples_leaf": 2,
    "max_depth": 10,
    "bootstrap": True,
    "class_weight": "balanced",
}


class RandomForest(Model):
    def __init__(self, x_metadata, **kwargs):
        name = "rf_lcld"
        self.model = DefaultRf(
            x_metadata=x_metadata, **rf_parameters, **kwargs
        )

        super(RandomForest, self).__init__(name)

    def fit(self, x, y):
        self.model.fit(x, y)

    def predict(self, x):
        return self.model.predict(x)

    def load(self, path):
        self.model.load(path)

    def save(self, path):
        self.model.save(path)


models = [("rf_lcld", RandomForest)]



import importlib
from typing import Any, Dict, Type, Union

from mlc.models.model import Model
from mlc.models.samples import load_model


def get_model_from_config(config: Dict[str, Any]) -> Type[Model]:
    name = config["name"]

    # If it is a known dataset
    if len(config.keys()) == 1:
        return load_model(name)

    # Else load and do
    models = importlib.import_module(config["source"]).models
    models_out = list(filter(lambda e: e[0] in [name], models))
    models_out = [e[1] for e in models_out]

    if len(models_out) != 1:
        raise NotImplementedError("At least one model is not available.")
    return models_out[0]


def get_model(config: Union[Dict[str, Any], str]) -> Type[Model]:
    if isinstance(config, str):
        config = {"name": config}
    return get_model_from_config(config)



from typing import Any, Dict, Union

from mlc.metrics.metric import Metric
from mlc.metrics.metrics import (
    PredClassificationMetric,
    ProbaClassificationMetric,
    RegressionMetric,
    str2pred_classification_metric,
    str2proba_classification_metric,
    str2regression_metric,
)


def create_metric(config: Union[str, Dict[str, Any]]) -> Metric:
    if isinstance(config, str):
        config = {"name": config}
    name = config.get("name")
    if name in str2pred_classification_metric:
        return PredClassificationMetric(name)
    elif name in str2proba_classification_metric:
        return ProbaClassificationMetric(name)
    elif name in str2regression_metric:
        return RegressionMetric(name)
    else:
        raise NotImplementedError(f"Metric {name} is not yet implemented.")



from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Type
import joblib

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset
from mlc.dataloaders.fast_dataloader import FastTensorDataLoader

from mlc.load_do_save import load_json, save_json
from mlc.models.model import Model
from mlc.models.torch_models import BaseModelTorch
from mlc.transformers.tab_scaler import TabScaler
from mlc.utils import parent_exists

import optuna
from joblib import Parallel, delayed

"""
    VIME: Extending the Success of Self- and
    Semi-supervised Learning to Tabular Domain
    (https://proceedings.neurips.cc/paper/2020/hash/7d97667a3e056acab9aaf653807b4a03-Abstract.html)

    Custom implementation using PyTorch.
    See the original implementation using
    Tensorflow: https://github.com/jsyoon0823/VIME
"""


class VIME(BaseModelTorch):
    def __init__(
        self,
        objective: str,
        x_metadata: pd.DataFrame,
        batch_size: int,
        epochs: int,
        early_stopping_rounds: int,
        num_classes: int,
        p_m: float = 0.3,
        K: int = 3,
        alpha: float = 2,
        beta: float = 1,
        val_batch_size: int = 100000,
        device="cpu",
        name="saint",
        scaler: Optional[TabScaler] = None,
        **kwargs: Any,
    ) -> None:

        self.objective = objective
        self.x_metadata = x_metadata
        self.batch_size = batch_size
        self.epochs = epochs
        self.early_stopping_rounds = early_stopping_rounds
        self.num_classes = num_classes
        self.device = device

        self.p_m = p_m
        self.K = K
        self.alpha = alpha
        self.beta = beta
        self.val_batch_size = val_batch_size

        if scaler is not None:
            self.scaler = TabScaler(num_scaler="min_max", one_hot_encode=False)
            self.scaler.fit_scaler_data(scaler.get_scaler_data())

        # Generate super call
        super().__init__(
            objective=objective,
            x_metadata=x_metadata,
            batch_size=batch_size,
            epochs=epochs,
            early_stopping_rounds=early_stopping_rounds,
            num_classes=num_classes,
            p_m=p_m,
            K=K,
            alpha=alpha,
            beta=beta,
            val_batch_size=val_batch_size,
            name=name,
            scaler=scaler,
            **kwargs,
        )

        print("On Device:", self.device)

        self.cat_idx = np.where(x_metadata["type"] == "cat")[0].tolist()
        self.num_features = x_metadata.shape[0]
        self.cat_idx = np.where(x_metadata["type"] == "cat")[0].tolist()
        self.num_idx = list(set(range(self.num_features)) - set(self.cat_idx))
        self.cat_dims = [
            (int(x_metadata.iloc[i]["max"]) + 1) for i in self.cat_idx
        ]

        self.model_self = VIMESelf(self.num_features).to(self.device)
        self.model_semi = VIMESemi(
            self.objective, self.num_features, self.num_classes
        ).to(self.device)

        self.encoder_layer = self.model_self.input_layer

        self.model = nn.Sequential(self.encoder_layer, self.model_semi)
        if hasattr(self, "scaler") and (self.scaler is not None):
            self.model = nn.Sequential(
                self.scaler.get_transorm_nn(), self.model
            )

        self.wrapper_model = self.model

        self.experiment = None

    def fit(
        self,
        X,
        y,
        X_val=None,
        y_val=None,
        custom_train_dataloader=None,
        custom_val_dataloader=None,
    ):

        X = self.scaler.transform(np.array(X, dtype=float))
        X_val = self.scaler.transform(np.array(X_val, dtype=float))

        X_unlab = np.concatenate([X, X_val], axis=0)

        self.fit_self(X_unlab, p_m=self.p_m, alpha=self.alpha)

        # if self.args.data_parallel:
        #     self.encoder_layer = self.model_self.module.input_layer
        # else:
        # self.encoder_layer = self.model_self.input_layer
        if self.device == "cuda":
            torch.cuda.synchronize()

        loss_history, val_loss_history = self.fit_semi(
            X,
            y,
            X,
            X_val,
            y_val,
            p_m=self.p_m,
            K=self.K,
            beta=self.beta,
            custom_train_dataloader=custom_train_dataloader,
            custom_val_dataloader=custom_val_dataloader,
            scaler=self.scaler
        )
        self.load_best_weights()
        return loss_history, val_loss_history

    def get_logits(self, x: torch.Tensor, with_grad: bool) -> torch.Tensor:
        if not with_grad:
            with torch.no_grad():
                # with_grad = True to avoid recursion
                return self.get_logits(x, with_grad=True)

        preds = self.model(x)
        if self.objective == "binary":
            preds = torch.sigmoid(preds)
        return preds

    def predict_helper(self, x, keep_grad=False):
        self.model.eval()

        x = np.array(x, dtype=float)
        x = torch.tensor(x).float().to(self.device)

        # test_dataset = TensorDataset(x)
        # test_loader = DataLoader(
        #     dataset=test_dataset,
        #     batch_size=self.val_batch_size,
        #     shuffle=False,
        #     num_workers=2,
        # )

        test_loader = FastTensorDataLoader(
            x, batch_size=self.val_batch_size, shuffle=False
        )

        predictions = []
        for batch_x in test_loader:
            preds = self.get_logits(batch_x[0], with_grad=keep_grad)
            predictions.append(preds.cpu())

        return np.concatenate(predictions)

    @staticmethod
    def define_trial_parameters(
        trial: optuna.Trial, trial_params: Dict[str, Any]
    ):
        params = {
            "p_m": trial.suggest_float("p_m", 0.1, 0.9),
            "alpha": trial.suggest_float("alpha", 0.1, 10),
            "K": trial.suggest_categorical("K", [2, 3, 5, 10, 15, 20]),
            "beta": trial.suggest_float("beta", 0.1, 10),
        }
        return params

    def fit_self(self, X, p_m=0.3, alpha=2):
        optimizer = optim.RMSprop(self.model_self.parameters(), lr=0.001)
        loss_func_mask = nn.BCELoss()
        loss_func_feat = nn.MSELoss()

        # X.to(self.device)
        # X = torch.tensor(X).float()
        # print(X.device)

        m_unlab = mask_generator(p_m, X)
        m_label, x_tilde = pretext_generator(m_unlab, X)
        X = torch.tensor(X).float()

        x_tilde = torch.tensor(x_tilde).float()
        m_label = torch.tensor(m_label).float()

        # If the dataset is relatively small, we can load it to the GPU
        if X.shape[1] < 10000:
            x_tilde = x_tilde.to(self.device)
            m_label = m_label.to(self.device)
            X = X.to(self.device)

        train_loader = FastTensorDataLoader(
            x_tilde,
            m_label,
            X,
            batch_size=self.batch_size,
            shuffle=True,
        )

        for epoch in range(10):
            for i, (batch_X, batch_mask, batch_feat) in enumerate(
                train_loader
            ):
                # print(F"Self supervised batch {i}")
                out_mask, out_feat = self.model_self(batch_X.to(self.device))

                loss_mask = loss_func_mask(
                    out_mask, batch_mask.to(self.device)
                )
                loss_feat = loss_func_feat(
                    out_feat, batch_feat.to(self.device)
                )

                loss = loss_mask + loss_feat * alpha

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

        print("Fitted encoder")

    def fit_semi(
        self,
        X,
        y,
        x_unlab,
        X_val=None,
        y_val=None,
        p_m=0.3,
        K=3,
        beta=1,
        custom_train_dataloader=None,
        custom_val_dataloader=None,
        scaler= None,
    ):
        X = torch.tensor(X).float()
        y = torch.tensor(y)
        x_unlab = torch.tensor(x_unlab).float()

        X_val = torch.tensor(X_val).float()
        y_val = torch.tensor(y_val)

        self.set_loss_y(y)

        y = y.float()
        y_val = y_val.float()

        if (y.ndim == 1) or (y.shape[1] == 1):
            y_train = torch.column_stack((1 - y, y))
        if (y_val.ndim == 1) or (y_val.shape[1] == 1):
            y_val = torch.column_stack((1 - y_val, y_val))

        loss_func_supervised = self.loss_func

        optimizer = optim.AdamW(self.model_semi.parameters())

        if custom_train_dataloader:
            tensors_x, tensors_y, tensors_x_unlab = X.to(self.device), y_train.to(self.device), x_unlab.to(self.device)
            if hasattr(custom_train_dataloader, "dataset"):
                custom_train_dataloader.dataset.tensors = tensors_x, tensors_y , tensors_x_unlab
            else:
                custom_train_dataloader.set_tensors(tensors_x, tensors_y, tensors_x_unlab)

            train_loader = custom_train_dataloader
            train_loader.scaler = scaler
            train_loader.transform_y = True
        else:
            if X.shape[1] < 10000:
                X = X.to(self.device)
                y_train = y_train.to(self.device)
                x_unlab = x_unlab.to(self.device)
            train_loader = FastTensorDataLoader(
                X,
                y_train,
                x_unlab,
                batch_size=self.batch_size,
                shuffle=True,
            )

        if custom_val_dataloader:
            custom_val_dataloader.dataset.tensors = (
                X_val.to(self.device),
                y_val.to(self.device),
            )
            val_loader = custom_val_dataloader
            val_loader.scaler = scaler
        else:
            if X.shape[1] < 10000:
                X_val = X_val.to(self.device)
                y_val = y_val.to(self.device)
            val_loader = FastTensorDataLoader(
                X_val,
                y_val,
                batch_size=self.batch_size,
                shuffle=False,
            )

        min_val_loss = float("inf")
        min_val_loss_idx = 0

        loss_history = []
        val_loss_history = []

        n_batch = int(np.ceil(X.shape[0] / self.batch_size))
        # Even if K=20 cutting in 5 shall fit in the GPU memory
        n_batch_per_precompute = int(np.ceil(n_batch / min(K, 5)))
        print("--------")
        print(X.shape)
        for epoch in range(self.epochs):
            print(K)

            for i, (batch_X, batch_y, batch_unlab) in enumerate(train_loader):

                i_relative_k = i % n_batch_per_precompute
                if i_relative_k == 0:
                    print(f"Precompute {i}")
                    x_for_k = x_unlab[
                        i
                        * self.batch_size : (i + n_batch_per_precompute)
                        * self.batch_size
                    ]
                    mask = torch_mask_generator(p_m, x_for_k.repeat(K, 1))
                    _, unlab_tmp = prepare_pretext_generator(
                        mask, x_for_k, self.batch_size, K
                    )
                    unlab_tmp = unlab_tmp.reshape(K, x_for_k.shape[0], -1)
                    print(unlab_tmp.shape)

                batch_X_encoded = self.encoder_layer(batch_X.to(self.device))
                y_hat = self.model_semi(batch_X_encoded)

                yv_hats = torch.empty(K, batch_X.shape[0], self.num_classes)
                for rep in range(K):
                    batch_unlab_tmp = unlab_tmp[rep][
                        (i_relative_k)
                        * self.batch_size : (i_relative_k + 1)
                        * self.batch_size
                    ]

                    batch_unlab_encoded = self.encoder_layer(
                        batch_unlab_tmp.to(self.device)
                    )
                    yv_hat = self.model_semi(batch_unlab_encoded)
                    yv_hats[rep] = yv_hat

                if (
                    self.objective == "regression"
                    or self.objective == "binary"
                ):
                    y_hat = y_hat.squeeze()

                y_loss = loss_func_supervised(y_hat, batch_y.to(self.device))
                yu_loss = torch.mean(torch.var(yv_hats, dim=0))
                loss = y_loss + beta * yu_loss
                loss_history.append(loss.item())
                if self.experiment is not None:
                    self.experiment.log_metric("train_loss", loss.item())

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            # Early Stopping
            val_loss = 0.0
            val_dim = 0
            with torch.no_grad():
                for val_i, (batch_val_X, batch_val_y) in enumerate(val_loader):
                    batch_val_X_encoded = self.encoder_layer(
                        batch_val_X.to(self.device)
                    )
                    y_hat = self.model_semi(batch_val_X_encoded)

                    if (
                        self.objective == "regression"
                        or self.objective == "binary"
                    ):
                        y_hat = y_hat.squeeze()

                    val_loss += loss_func_supervised(
                        y_hat, batch_val_y.to(self.device)
                    ).item()
                    val_dim += 1

            val_loss /= val_dim
            val_loss_history.append(val_loss)

            if self.experiment is not None:
                self.experiment.log_metric("validation_loss", val_loss)

            print("Epoch %d, Val Loss: %.5f" % (epoch, val_loss))

            if val_loss < min_val_loss:
                min_val_loss = val_loss
                min_val_loss_idx = epoch

                self.save_best_weights()

                # self.save_model(filename_extension="best", directory="tmp")

            if min_val_loss_idx + self.early_stopping_rounds < epoch:
                print("Early stopping applies.")
                break

        return loss_history, val_loss_history

    def save(self, path: str) -> None:
        kwargs_save = deepcopy(self.constructor_kwargs)
        kwargs_save.pop("x_metadata")

        if "scaler" in kwargs_save:
            kwargs_save.pop("scaler")
        args_path = f"{path}/args.json"
        save_json(kwargs_save, parent_exists(args_path))

        weigths_path = f"{path}/weights.pt"
        torch.save(self.model_self.state_dict(), parent_exists(weigths_path))

        semi_weights_path = f"{path}/semi_weights.pt"
        torch.save(
            self.model_semi.state_dict(), parent_exists(semi_weights_path)
        )

    @classmethod
    def load_class(cls, path: str, **kwargs: Dict[str, Any]) -> BaseModelTorch:

        args_path = f"{path}/args.json"
        weigths_path = f"{path}/weights.pt"
        semi_weights_path = f"{path}/semi_weights.pt"

        load_kwargs = load_json(args_path)
        args = {**load_kwargs, **kwargs}
        model = cls(**args)
        model.model_self.load_state_dict(torch.load(weigths_path))
        model.model_semi.load_state_dict(torch.load(semi_weights_path))

        return model

    def save_best_weights(self) -> None:
        for path, model in zip(
            [f"./tmp/{self.uuid}_best.pt", f"./tmp/{self.uuid}_semi_best.pt"],
            [self.model_self, self.model_semi],
        ):
            torch.save(model.state_dict(), parent_exists(path))

    def load_best_weights(self) -> None:
        for path, model in zip(
            [f"./tmp/{self.uuid}_best.pt", f"./tmp/{self.uuid}_semi_best.pt"],
            [self.model_self, self.model_semi],
        ):
            model.load_state_dict(torch.load(path))

    def get_model_size(self):
        self_size = sum(
            t.numel() for t in self.model_self.parameters() if t.requires_grad
        )
        semi_size = sum(
            t.numel() for t in self.model_semi.parameters() if t.requires_grad
        )
        return self_size + semi_size


class VIMESelf(nn.Module):
    def __init__(self, input_dim):
        super().__init__()

        self.input_layer = nn.Linear(input_dim, input_dim)

        self.mask_layer = nn.Linear(input_dim, input_dim)
        self.feat_layer = nn.Linear(input_dim, input_dim)

    def forward(self, x):
        x = F.relu(self.input_layer(x))

        out_mask = torch.sigmoid(self.mask_layer(x))
        out_feat = torch.sigmoid(self.feat_layer(x))

        return out_mask, out_feat


class VIMESemi(nn.Module):
    def __init__(
        self, objective, input_dim, output_dim, hidden_dim=100, n_layers=5
    ):
        super().__init__()
        self.objective = objective

        self.input_layer = nn.Linear(input_dim, hidden_dim)

        self.layers = nn.ModuleList()
        self.layers.extend(
            [nn.Linear(hidden_dim, hidden_dim) for _ in range(n_layers - 1)]
        )

        self.output_layer = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        x = F.relu(self.input_layer(x))

        for layer in self.layers:
            x = F.relu(layer(x))

        out = self.output_layer(x)

        if self.objective == "classification":
            out = F.softmax(out, dim=1)

        return out


"""
    VIME code copied from: https://github.com/jsyoon0823/VIME
"""


def mask_generator(p_m, x):
    mask = np.random.binomial(1, p_m, x.shape)
    return mask


def torch_mask_generator(p_m, x):
    return torch.bernoulli(torch.full(x.shape, p_m).to(x.device))


def pretext_generator(m, x):
    # Parameters
    no, dim = x.shape
    # Convert the input numpy array to a PyTorch tensor
    # x = torch.tensor(x)

    # Randomly (and column-wise) shuffle data
    # Randomly (and column-wise) shuffle data

    x_bar = x.copy()

    np.apply_along_axis(np.random.shuffle, 0, x_bar)
    # for i in range(arr.shape[1]):
    #     np.random.shuffle(arr[:, i])
    # x_bar = arr

    # x_bar = np.zeros([no, dim])
    # for i in range(dim):
    #     idx = np.random.permutation(no)
    #     x_bar[:, i] = x[idx, i]

    # Corrupt samples
    # x_bar = torch.tensor(x_bar).float()
    x_tilde = x * (1 - m) + x_bar * m
    # Define new mask matrix
    m_new = 1 * (x != x_tilde)

    return m_new, x_tilde


def prepare_permut(x):
    x_bar = x.copy()
    np.apply_along_axis(np.random.shuffle, 0, x_bar)
    return x_bar


def prepare_pretext_generator(m, x, batch_size, K):

    n_batch = np.ceil(x.shape[0] / batch_size)
    x_batch = np.array_split(x.cpu().numpy(), n_batch) * K

    x_bar = np.concatenate(
        Parallel(n_jobs=max(10, joblib.cpu_count()))(
            delayed(prepare_permut)(e) for e in x_batch
        )
    )

    x_bar = torch.tensor(x_bar).float().to(x.device)

    x = x.repeat(K, 1)
    x_tilde = x * (1 - m) + x_bar * m
    m_new = (1 * (x != x_tilde)).float()

    return m_new, x_tilde


models: List[Tuple[str, Type[Model]]] = [("vime", VIME)]



import os

import numpy as np

from .lib import Transformations, load_json, read_pure_data, transform_dataset, Dataset
from .lib import change_val as change_valF
from .modules import MLPDiffusion, ResNetDiffusion


def get_model(
        model_name,
        model_params,
        n_num_features,
        category_sizes
):
    print(model_name)
    if model_name == 'mlp':
        model = MLPDiffusion(**model_params)
    elif model_name == 'resnet':
        model = ResNetDiffusion(**model_params)
    else:
        raise "Unknown model!"
    return model


def update_ema(target_params, source_params, rate=0.999):
    """
    Update target parameters to be closer to those of source parameters using
    an exponential moving average.
    :param target_params: the target parameter sequence.
    :param source_params: the source parameter sequence.
    :param rate: the EMA rate (closer to 1 means slower).
    """
    for targ, src in zip(target_params, source_params):
        targ.detach().mul_(rate).add_(src.detach(), alpha=1 - rate)


def concat_y_to_X(X, y):
    if X is None:
        return y.reshape(-1, 1)
    return np.concatenate([y.reshape(-1, 1), X], axis=1)


def make_dataset(
        data_path: str,
        T: Transformations,
        num_classes: int,
        is_y_cond: bool,
        change_val: bool
):
    # classification
    if num_classes > 0:
        X_cat = {} if os.path.exists(os.path.join(data_path, 'X_cat_train.npy')) or not is_y_cond else None
        X_num = {} if os.path.exists(os.path.join(data_path, 'X_num_train.npy')) else None
        y = {}

        for split in ['train', 'val', 'test']:
            X_num_t, X_cat_t, y_t = read_pure_data(data_path, split)
            if X_num is not None:
                X_num[split] = X_num_t
            if not is_y_cond:
                X_cat_t = concat_y_to_X(X_cat_t, y_t)
            if X_cat is not None:
                X_cat[split] = X_cat_t
            y[split] = y_t
    else:
        # regression
        X_cat = {} if os.path.exists(os.path.join(data_path, 'X_cat_train.npy')) else None
        X_num = {} if os.path.exists(os.path.join(data_path, 'X_num_train.npy')) or not is_y_cond else None
        y = {}

        for split in ['train', 'val', 'test']:
            X_num_t, X_cat_t, y_t = read_pure_data(data_path, split)
            if not is_y_cond:
                X_num_t = concat_y_to_X(X_num_t, y_t)
            if X_num is not None:
                X_num[split] = X_num_t
            if X_cat is not None:
                X_cat[split] = X_cat_t
            y[split] = y_t

    info = load_json(os.path.join(data_path, 'info.json'))

    D = Dataset(
        X_num,
        X_cat,
        y,
        y_info={},
        task_type=lib.TaskType(info['task_type']),
        n_classes=info.get('n_classes')
    )

    if change_val:
        D = change_valF(D)

    return transform_dataset(D, T, None)



from typing import Any, Dict, Union

import joblib
import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder, StandardScaler

from mlc.transformers.tokennizer import TokenizerTransformer
from mlc.transformers.transformer import Transformer


class TabTransformer(Transformer):
    def __init__(
        self,
        metadata: pd.DataFrame,
        scale: bool,
        one_hot_encode: bool,
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(
            name="tab_transformer",
            metadata=metadata,
            scale=scale,
            one_hot_encode=one_hot_encode,
            **kwargs,
        )
        cat_index = metadata["type"] == "cat"
        cat_features = metadata[cat_index]["feature"]
        num_index = (metadata["type"] == "int") | (metadata["type"] == "real")
        num_features = metadata[num_index]["feature"]
        text_index = metadata["type"] == "text"
        text_features = metadata[text_index]["feature"]

        cat_min = metadata[cat_index]["min"].to_list()
        cat_max = metadata[cat_index]["max"].to_list()
        cat_range = [
            np.arange(int(cat_min[i]), int(cat_max[i]) + 1)
            for i in range(len(cat_min))
        ]
        transformers = []
        if scale:
            transformers.append(("num", StandardScaler(), num_features))
        if one_hot_encode:
            transformers.append(
                (
                    "cat",
                    OneHotEncoder(
                        sparse=False,
                        handle_unknown="ignore",
                        drop="if_binary",
                        categories=cat_range,
                    ),
                    cat_features,
                )
            )

        if len(metadata[text_index]["max"]) > 0:
            self.max_token_length = int(metadata[text_index]["max"].max())
            transformers.append(
                (
                    "text",
                    TokenizerTransformer(self.max_token_length),
                    text_features,
                )
            )

        self.transformer = ColumnTransformer(
            transformers=transformers,
            sparse_threshold=0,
            remainder="passthrough",
            n_jobs=-1,
        )

    def fit(
        self,
        x: Union[npt.NDArray[np.float_], pd.DataFrame],
        y: Union[npt.NDArray[np.int_], pd.Series] = None,
    ) -> Transformer:
        self.transformer.fit(x, y)
        return self

    def transform(
        self, x: Union[npt.NDArray[np.float_], pd.DataFrame]
    ) -> Union[npt.NDArray[np.float_], pd.DataFrame]:
        return self.transformer.transform(x)

    def load(self, path: str) -> None:
        text_pending_idx = []
        for i, e in enumerate(self.transformer.transformers):
            if e[0] == "text":
                text_pending_idx.append(i)

        text_pending = [
            self.transformer.transformers[i] for i in text_pending_idx
        ]
        self.transformer = joblib.load(path)
        for i in range(len(text_pending_idx)):
            self.transformer.transformers[text_pending_idx[i]] = text_pending[
                i
            ]
            self.transformer.transformers_[text_pending_idx[i]] = text_pending[
                i
            ]

    def save(self, path: str) -> None:
        text_pending_idx = []
        for i, e in enumerate(self.transformer.transformers):
            if e[0] == "text":
                text_pending_idx.append(i)

        text_pending = [
            self.transformer.transformers[i] for i in text_pending_idx
        ]
        for i in text_pending_idx:
            self.transformer.transformers[i] = None
            self.transformer.transformers_[i] = None

        joblib.dump(self.transformer, path)

        for i in range(len(text_pending_idx)):
            self.transformer.transformers[text_pending_idx[i]] = text_pending[
                i
            ]
            self.transformer.transformers_[text_pending_idx[i]] = text_pending[
                i
            ]



import re
from typing import Dict, List

import numpy as np
import numpy.typing as npt
from sklearn.model_selection import train_test_split

from mlc.constraints.relation_constraint import (
    BaseRelationConstraint,
    Constant,
    Count,
    Feature,
    Log,
    ManySum,
)
from mlc.datasets.dataset import (
    CsvDataSource,
    Dataset,
    DataSource,
    DefaultIndexSorter,
    DownloadFileDataSource,
    Hdf5DataSource,
    Splitter,
    Task,
)


class MalwareSplitter(Splitter):
    def get_splits(self, dataset: Dataset) -> Dict[str, npt.NDArray[np.int_]]:
        _, y = dataset.get_x_y()
        i = np.arange(len(y))
        i_train, i_test = i[:12308], i[12308:]
        i_val, i_test = train_test_split(
            i_test, test_size=0.5, random_state=42
        )
        return {"train": i_train, "val": i_val, "test": i_test}


def get_relation_constraints(
    metadata: DataSource,
) -> List[BaseRelationConstraint]:

    features = metadata.load_data()["feature"].to_list()

    def regex_features(features_l: List[str], pattern: str) -> List[Feature]:
        p = re.compile(pattern)
        out = list(filter(lambda el: p.match(el), features_l))
        return [Feature(o) for o in out]

    features_g1 = regex_features(features, "^pesection_\d+_name$")
    g1_right = [f == Constant(832) for f in features_g1]
    g1 = Feature("header_NumberOfSections") == Count(g1_right, inverse=True)

    g2 = Feature("header_FileAlignment") <= Feature("header_SectionAlignment")

    g3 = Log(Feature("header_FileAlignment")) / Log(Constant(2.0))
    g3 = g3 % Constant(1.0)
    g3 = g3 == Constant(0.0)

    g4 = ManySum(regex_features(features, "^imp_.*$")) <= Feature(
        "api_import_nb"
    )

    g5 = ManySum(regex_features(features, "^.*\.dll$")) <= Feature(
        "dll_import_nb"
    )

    g6 = ManySum(regex_features(features, "^freq_byte_\d+$")) == Constant(1)

    freq = regex_features(features, "^freq_byte_\d+$")
    log_freq = [
        (f * (Log(f, safe_value=Constant(0.0)) / Log(Constant(2.0))))
        for f in freq
    ]
    g7 = Feature("generic_fileEntropy") == (Constant(0) - ManySum(log_freq))

    return [g1, g2, g3, g4, g5, g6, g7]


def create_dataset() -> Dataset:
    data_source = DownloadFileDataSource(
        url="https://uniluxembourg-my.sharepoint.com/:u:/g/personal/"
        "thibault_simonetto_uni_lu/"
        "EWpXWGxq_WxBtCJ5o1mr0fcBGcJYFrYchZD8esq-91yfJw?download=1",
        file_data_source=Hdf5DataSource(
            path="./data/mlc/malware/malware.hdf5"
        ),
    )
    metadata_source = DownloadFileDataSource(
        url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
        "thibault_simonetto_uni_lu/"
        "Ef-mGuyoTZtBqnUe6JA9HK4BiqyxHrVYgO7ssJ0Qe6T_AQ?download=1",
        file_data_source=CsvDataSource(
            path="./data/mlc/malware/malware_metadata.csv"
        ),
    )
    tasks = [
        Task(
            name="malicious",
            task_type="classification",
            evaluation_metric="mcc",
        )
    ]
    sorter = DefaultIndexSorter()
    splitter = MalwareSplitter()
    relation_constraints = get_relation_constraints(metadata_source)

    malware = Dataset(
        name="malware",
        data_source=data_source,
        metadata_source=metadata_source,
        tasks=tasks,
        sorter=sorter,
        splitter=splitter,
        relation_constraints=relation_constraints,
    )
    return malware


datasets = [
    {
        "name": "malware",
        "fun_create": create_dataset,
    }
]



import numpy as np
import numpy.typing as npt
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    confusion_matrix,
    f1_score,
    matthews_corrcoef,
    mean_squared_error,
    roc_auc_score,
)

from mlc.metrics.metric import Metric


def transaction_cost(
    y: npt.NDArray[np.int_], y_s: npt.NDArray[np.int_]
) -> float:
    (tn, fp, fn, tp) = confusion_matrix(y, y_s)
    return (((tp + fp) * 5.0) + (fn * 60.0) + (tn * 0)) / 60


str2pred_classification_metric = {
    "f1": f1_score,
    "mcc": matthews_corrcoef,
    "error_rate": lambda y, y_s: 1 - accuracy_score(y, y_s),
    "class_error": lambda y, y_s: 1 - (y == y_s).astype(int),
    "accuracy": lambda y, y_s: accuracy_score(y, y_s),
    "confusion_matrix": lambda y, y_s: confusion_matrix(y, y_s),
    "balanced_accuracy": lambda y, y_s: balanced_accuracy_score(y, y_s),
    "transaction_cost": transaction_cost,
}
str2proba_classification_metric = {
    "auc": lambda y, y_s: roc_auc_score(
        y,
        y_s,
    ),
    "proba_error": lambda y, y_s: 1 - y_s[np.arange(len(y)), y],
    "y_scores": lambda y, y_s: y_s,
}

str2regression_metric = {
    "rmse": lambda y, y_s: mean_squared_error(y, y_s, squared=True)
}


class RecordedMetric(object):
    def __init__(self):
        self.reset()

    def mean(self):
        return np.mean(self.values)

    def std(self):
        return np.std(self.values)

    def reset(self):
        self.values = []

    def last(self):
        return self.values[-1]


class PredClassificationMetric(Metric, RecordedMetric):
    def __init__(self, metric_name: str):
        self.metric_name = metric_name
        self.values = []

    def compute(
        self, y_true: npt.NDArray[np.generic], y_score: npt.NDArray[np.generic]
    ) -> npt.NDArray[np.generic]:
        val = str2pred_classification_metric[self.metric_name](y_true, y_score)

        self.values.append(val)
        return val


class ProbaClassificationMetric(Metric, RecordedMetric):
    def __init__(self, metric_name: str):
        self.metric_name = metric_name
        self.values = []

    def compute(
        self, y_true: npt.NDArray[np.generic], y_score: npt.NDArray[np.generic]
    ) -> npt.NDArray[np.generic]:

        val = str2proba_classification_metric[self.metric_name](
            y_true, y_score
        )

        self.values.append(val)
        return val


class RegressionMetric(Metric, RecordedMetric):
    def __init__(self, metric_name: str):
        self.metric_name = metric_name
        self.values = []

    def compute(
        self, y_true: npt.NDArray[np.generic], y_score: npt.NDArray[np.generic]
    ) -> npt.NDArray[np.generic]:
        val = str2regression_metric[self.metric_name](y_true, y_score)

        self.values.append(val)
        return val



import typing
from abc import abstractmethod
from typing import Any, Callable, Dict, Union

import numpy as np
import numpy.typing as npt
import torch

from mlc.constraints.relation_constraint import (
    AndConstraint,
    BaseRelationConstraint,
    Constant,
    ConstraintsNode,
    Count,
    EqualConstraint,
    Feature,
    LessConstraint,
    LessEqualConstraint,
    Log,
    ManySum,
    MathOperation,
    OrConstraint,
    SafeDivision,
    Value,
)
from mlc.constraints.utils import get_feature_index

EPS: npt.NDArray[Any] = np.array(0.000001)


class ConstraintsVisitor:
    """Abstract Visitor Class"""

    @abstractmethod
    def visit(self, item: ConstraintsNode) -> Any:
        pass

    @abstractmethod
    def execute(self) -> Any:
        pass


class NumpyConstraintsVisitor(ConstraintsVisitor):

    str_operator_to_result = {
        "+": lambda left, right: left + right,
        "-": lambda left, right: left - right,
        "*": lambda left, right: left * right,
        "/": lambda left, right: left / right,
        "**": lambda left, right: left**right,
        "%": lambda left, right: left % right,
    }

    def __init__(
        self,
        constraint: Union[BaseRelationConstraint, Value],
        x: npt.NDArray[Any],
        feature_names: npt.ArrayLike = None,
    ) -> None:
        self.constraint = constraint
        self.x = x
        self.feature_names = feature_names

    @staticmethod
    def get_zeros_np(
        operands: typing.List[npt.NDArray[Any]],
    ) -> npt.NDArray[Any]:
        i = np.argmax([op.ndim for op in operands])
        return np.zeros(operands[i].shape, dtype=operands[i].dtype)

    def visit(self, constraint_node: ConstraintsNode) -> npt.NDArray[Any]:

        # ------------ Values
        if isinstance(constraint_node, Constant):
            return np.array(constraint_node.constant)

        elif isinstance(constraint_node, Feature):

            feature_index = get_feature_index(
                self.feature_names, constraint_node.feature_id
            )
            return self.x[:, feature_index]

        elif isinstance(constraint_node, MathOperation):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            operator = constraint_node.operator
            if operator in self.str_operator_to_result:
                return self.str_operator_to_result[operator](
                    left_operand, right_operand
                )
            else:
                raise NotImplementedError

        elif isinstance(constraint_node, SafeDivision):
            dividend = constraint_node.dividend.accept(self)
            divisor = constraint_node.divisor.accept(self)
            fill_value = constraint_node.fill_value.accept(self)
            return np.divide(
                dividend,
                divisor,
                out=np.full_like(dividend, fill_value),
                where=divisor != 0,
            )

        elif isinstance(constraint_node, Log):
            operand = constraint_node.operand.accept(self)
            if constraint_node.safe_value is not None:
                safe_value = constraint_node.safe_value.accept(self)
                return np.log(
                    operand,
                    out=np.full_like(operand, fill_value=safe_value),
                    where=(operand > 0),
                )
            return np.log(operand)

        elif isinstance(constraint_node, ManySum):
            operands = [e.accept(self) for e in constraint_node.operands]
            return np.sum(operands, axis=0)

        # ------------ Constraints

        # ------ Binary
        elif isinstance(constraint_node, OrConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            return np.min(operands, axis=0)

        elif isinstance(constraint_node, AndConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            return np.sum(operands, axis=0)

        # ------ Comparison
        elif isinstance(constraint_node, LessEqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            zeros = self.get_zeros_np([left_operand, right_operand])
            return np.max([zeros, (left_operand - right_operand)], axis=0)

        elif isinstance(constraint_node, LessConstraint):
            left_operand = constraint_node.left_operand.accept(self) + EPS
            right_operand = constraint_node.right_operand.accept(self)
            zeros = self.get_zeros_np([left_operand, right_operand])
            return np.max([zeros, (left_operand - right_operand)], axis=0)

        elif isinstance(constraint_node, EqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            return np.abs(left_operand - right_operand)

        # ------ Extension

        elif isinstance(constraint_node, Count):
            operands = [e.accept(self) for e in constraint_node.operands]
            if constraint_node.inverse:
                operands = [(op != 0).astype(float) for op in operands]
            else:
                operands = [(op == 0).astype(float) for op in operands]

            return np.sum(np.array(operands), axis=0)

        else:
            raise NotImplementedError

    def execute(self) -> npt.NDArray[Any]:
        return self.constraint.accept(self)


class NumpyConstraintsExecutor:
    def __init__(
        self,
        constraint: Union[BaseRelationConstraint, Value],
        feature_names: npt.ArrayLike = None,
    ):
        self.constraint = constraint
        self.feature_names = feature_names

    def execute(self, x: npt.NDArray[Any]) -> npt.NDArray[Any]:
        visitor = NumpyConstraintsVisitor(
            self.constraint, x, self.feature_names
        )
        return visitor.execute()


class TensorFlowConstraintsVisitor(ConstraintsVisitor):

    import tensorflow as tf

    def __init__(
        self,
        constraint: BaseRelationConstraint,
        x: "tf.Tensor",
        feature_names: npt.ArrayLike = None,
    ):
        self.constraint = constraint
        self.x = x
        self.feature_names = feature_names

    @staticmethod
    def get_zeros_tf(operands: typing.List["tf.Tensor"]) -> "tf.Tensor":
        import tensorflow as tf

        i = np.argmax([op.ndim for op in operands])
        return tf.zeros(operands[i].shape, dtype=operands[i].dtype)

    @staticmethod
    def str_operator_to_result_f() -> Dict[
        str, Callable[["tf.Tensor", "tf.Tensor"], "tf.Tensor"]
    ]:
        import tensorflow as tf

        return {
            "+": lambda left, right: left + right,
            "-": lambda left, right: left - right,
            "*": lambda left, right: left * right,
            "/": lambda left, right: left / right,
            "**": lambda left, right: tf.math.pow(left, right),
        }

    def visit(self, constraint_node: ConstraintsNode) -> "tf.Tensor":
        import tensorflow as tf

        # ------------ Values
        if isinstance(constraint_node, Constant):
            return tf.constant(constraint_node.constant, dtype=tf.float32)

        elif isinstance(constraint_node, Feature):
            feature_index = get_feature_index(
                self.feature_names, constraint_node.feature_id
            )
            return self.x[:, feature_index]

        elif isinstance(constraint_node, MathOperation):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            operator = constraint_node.operator
            if operator in self.str_operator_to_result_f():
                return self.str_operator_to_result_f()[operator](
                    left_operand, right_operand
                )
            else:
                raise NotImplementedError

        # ------------ Constraints

        # ------ Binary
        elif isinstance(constraint_node, OrConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            local_min = operands[0]
            for i in range(1, len(operands)):
                local_min = tf.minimum(local_min, operands[i])
            return local_min

        elif isinstance(constraint_node, AndConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            local_sum = operands[0]
            for i in range(1, len(operands)):
                local_sum = local_sum + operands[i]
            return local_sum

        # ------ Comparison
        elif isinstance(constraint_node, LessEqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            zeros = self.get_zeros_tf([left_operand, right_operand])
            return tf.maximum(zeros, (left_operand - right_operand))

        elif isinstance(constraint_node, LessConstraint):
            left_operand = constraint_node.left_operand.accept(self) + EPS
            right_operand = constraint_node.right_operand.accept(self)
            zeros = self.get_zeros_tf([left_operand, right_operand])
            return tf.maximum(zeros, (left_operand - right_operand))

        elif isinstance(constraint_node, EqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            return tf.abs(left_operand - right_operand)

        else:
            raise NotImplementedError

    def execute(self) -> "tf.Tensor":
        return self.constraint.accept(self)


class TensorFlowConstraintsExecutor:

    import tensorflow as tf

    def __init__(
        self,
        constraint: BaseRelationConstraint,
        feature_names: npt.ArrayLike = None,
    ):
        self.constraint = constraint
        self.feature_names = feature_names

    def execute(self, x: "tf.Tensor") -> "tf.Tensor":
        visitor = TensorFlowConstraintsVisitor(
            self.constraint, x, self.feature_names
        )
        return visitor.execute()


class TextConstraintsVisitor(ConstraintsVisitor):

    str_operator_to_result = {
        "+": lambda left, right: left + right,
        "-": lambda left, right: left - right,
        "*": lambda left, right: left * right,
        "/": lambda left, right: left / right,
        "**": lambda left, right: left**right,
    }

    def __init__(
        self,
        constraint: BaseRelationConstraint,
    ) -> None:
        self.constraint = constraint

    def visit(self, constraint_node: ConstraintsNode) -> str:
        # ------------ Values
        if isinstance(constraint_node, Constant):
            return str(constraint_node.constant)

        elif isinstance(constraint_node, Feature):
            return f"F({str(constraint_node.feature_id)})"

        elif isinstance(constraint_node, MathOperation):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            operator = constraint_node.operator
            return f"({left_operand} {operator} {right_operand})"

        elif isinstance(constraint_node, SafeDivision):
            dividend = constraint_node.dividend.accept(self)
            divisor = constraint_node.divisor.accept(self)
            fill_value = constraint_node.fill_value.accept(self)
            return f"SafeDiv({dividend}, {divisor}, {fill_value})"

        # ------------ Constraints

        # ------ Binary
        elif isinstance(constraint_node, OrConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            out = " OR ".join(operands)
            return f"({out})"

        elif isinstance(constraint_node, AndConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            out = " AND ".join(operands)
            return f"({out})"

        # ------ Comparison
        elif isinstance(constraint_node, LessEqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            return f"({left_operand} <= {right_operand})"

        elif isinstance(constraint_node, LessConstraint):
            left_operand = constraint_node.left_operand.accept(self) + EPS
            right_operand = constraint_node.right_operand.accept(self)
            return f"({left_operand} < {right_operand})"

        elif isinstance(constraint_node, EqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            return f"({left_operand} == {right_operand})"

        else:
            raise NotImplementedError

    def execute(self) -> str:
        return self.constraint.accept(self)


class TextConstraintsExecutor:
    def __init__(
        self,
        constraint: BaseRelationConstraint,
    ):
        self.constraint = constraint

    def execute(self) -> str:
        visitor = TextConstraintsVisitor(self.constraint)
        return visitor.execute()


class PytorchConstraintsVisitor(ConstraintsVisitor):

    str_operator_to_result = {
        "+": lambda left, right: left + right,
        "-": lambda left, right: left - right,
        "*": lambda left, right: left * right,
        "/": lambda left, right: left / right,
        "**": lambda left, right: left**right,
        "%": lambda left, right: left % right,
    }

    def __init__(
        self,
        constraint: BaseRelationConstraint,
        x: "torch.Tensor",
        feature_names: npt.ArrayLike = None,
    ) -> None:
        self.constraint = constraint
        self.x = x
        self.feature_names = feature_names

    @staticmethod
    def get_zeros_np(
        operands: typing.List["torch.Tensor"],
    ) -> "torch.Tensor":
        i = np.argmax([op.ndim for op in operands])
        return torch.zeros(operands[i].shape, dtype=operands[i].dtype)

    def visit(self, constraint_node: ConstraintsNode) -> "torch.Tensor":

        # ------------ Values
        if isinstance(constraint_node, Constant):
            return torch.from_numpy(np.array(constraint_node.constant))

        elif isinstance(constraint_node, Feature):
            feature_index = get_feature_index(
                self.feature_names, constraint_node.feature_id
            )
            return self.x[:, feature_index]

        elif isinstance(constraint_node, MathOperation):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            operator = constraint_node.operator
            if operator in self.str_operator_to_result:
                return self.str_operator_to_result[operator](
                    left_operand, right_operand
                )
            else:
                raise NotImplementedError

        elif isinstance(constraint_node, SafeDivision):
            dividend = constraint_node.dividend.accept(self)
            divisor = constraint_node.divisor.accept(self)
            fill_value = constraint_node.fill_value.accept(self)
            return torch.where(
                divisor != 0, torch.div(dividend, divisor), fill_value
            )

        elif isinstance(constraint_node, Log):
            operand = constraint_node.operand.accept(self)
            if constraint_node.safe_value is not None:
                safe_value = constraint_node.safe_value.accept(self)
                return torch.where(operand > 0, torch.log(operand), safe_value)

            return torch.log(operand)

        elif isinstance(constraint_node, ManySum):
            operands = [e.accept(self) for e in constraint_node.operands]
            return torch.sum(torch.stack(operands), dim=0)

        # ------------ Constraints

        # ------ Binary
        elif isinstance(constraint_node, OrConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            return torch.min(torch.stack(operands), dim=0).values

        elif isinstance(constraint_node, AndConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            local_sum = operands[0]
            for i in range(1, len(operands)):
                local_sum = local_sum + operands[i]
            return local_sum

        # ------ Comparison
        elif isinstance(constraint_node, LessEqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            zeros = self.get_zeros_np([left_operand, right_operand])
            my_sub = left_operand - right_operand
            my_test = torch.stack([zeros, my_sub])
            return torch.max(my_test, dim=0)[0]

        elif isinstance(constraint_node, LessConstraint):
            left_operand = constraint_node.left_operand.accept(self) + EPS
            right_operand = constraint_node.right_operand.accept(self)
            zeros = self.get_zeros_np([left_operand, right_operand])
            return torch.max(
                torch.stack([zeros, (left_operand - right_operand)]), dim=0
            )

        elif isinstance(constraint_node, EqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            return torch.abs(left_operand - right_operand)

            # ------ Extension

        elif isinstance(constraint_node, Count):
            operands = [e.accept(self) for e in constraint_node.operands]
            if constraint_node.inverse:
                operands = [(op != 0).float() for op in operands]
            else:
                operands = [(op == 0).float() for op in operands]
            return torch.sum(torch.stack(operands), dim=0)

        else:
            raise NotImplementedError

    def execute(self) -> "torch.Tensor":
        return self.constraint.accept(self)


class PytorchConstraintsExecutor:
    def __init__(
        self,
        constraint: BaseRelationConstraint,
        feature_names: npt.ArrayLike = None,
    ):
        self.constraint = constraint
        self.feature_names = feature_names

    def execute(self, x: "torch.Tensor") -> "torch.Tensor":
        visitor = PytorchConstraintsVisitor(
            self.constraint, x, self.feature_names
        )
        return visitor.execute()



import numpy as np
import torch

from mlc.dataloaders.dataloader import BaseDataLoader
from mlc.dataloaders.fast_dataloader import FastTensorDataLoader




class DefaultDataLoader(BaseDataLoader):
    def __init__(self, scaler=None, model=None,
                 seed: int = 42, subset: float = 1, move_all_to_model_device: bool = True, **kwargs) -> None:

        self.subset = subset
        self.seed = seed
        self.scaler = scaler
        if model is not None:
            self.device = model.device
            self.model = model.wrapper_model
        self.move_all_to_model_device = move_all_to_model_device

    def get_dataloader(
            self, x: np.ndarray, y: np.ndarray, train: bool, batch_size: int
    ) -> FastTensorDataLoader:

        x = torch.Tensor(x).float()
        y = torch.Tensor(y).long()

        if self.subset < 1:
            dataset_size = len(x)
            split = int(np.floor(self.subset * dataset_size))
            x = x[0:split]
            y = y[0:split]

        if self.move_all_to_model_device:
            x = x.to(self.device)
            y = y.to(self.device)

        # Create dataloader
        if train:

            self.dataloader =  FastTensorDataLoader(
                x, y,
                batch_size=batch_size,
                shuffle=True
            )

        else:
            self.dataloader =  FastTensorDataLoader(
                x, y,
                batch_size=batch_size,
                shuffle=False
            )
        self.dataloader.scaler = None
        self.dataloader.transform_y = None

        return self.dataloader


import __main__
import argparse
import atexit
import enum
import json
import os
import pickle
import shutil
import sys
import time
import uuid
from copy import deepcopy
from dataclasses import asdict, fields, is_dataclass
from pathlib import Path
from pprint import pprint
from typing import Any, Callable, List, Dict, Type, Optional, Tuple, TypeVar, Union, cast, get_args, get_origin

import numpy as np
import tomli
import tomli_w
import torch
import zero

from . import env

RawConfig = Dict[str, Any]
Report = Dict[str, Any]
T = TypeVar('T')


class Part(enum.Enum):
    TRAIN = 'train'
    VAL = 'val'
    TEST = 'test'

    def __str__(self) -> str:
        return self.value


class TaskType(enum.Enum):
    BINCLASS = 'binclass'
    MULTICLASS = 'multiclass'
    REGRESSION = 'regression'

    def __str__(self) -> str:
        return self.value


class Timer(zero.Timer):
    @classmethod
    def launch(cls) -> 'Timer':
        timer = cls()
        timer.run()
        return timer


def update_training_log(training_log, data, metrics):
    def _update(log_part, data_part):
        for k, v in data_part.items():
            if isinstance(v, dict):
                _update(log_part.setdefault(k, {}), v)
            elif isinstance(v, list):
                log_part.setdefault(k, []).extend(v)
            else:
                log_part.setdefault(k, []).append(v)

    _update(training_log, data)
    transposed_metrics = {}
    for part, part_metrics in metrics.items():
        for metric_name, value in part_metrics.items():
            transposed_metrics.setdefault(metric_name, {})[part] = value
    _update(training_log, transposed_metrics)


def raise_unknown(unknown_what: str, unknown_value: Any):
    raise ValueError(f'Unknown {unknown_what}: {unknown_value}')


def _replace(data, condition, value):
    def do(x):
        if isinstance(x, dict):
            return {k: do(v) for k, v in x.items()}
        elif isinstance(x, list):
            return [do(y) for y in x]
        else:
            return value if condition(x) else x

    return do(data)


_CONFIG_NONE = '__none__'


def unpack_config(config: RawConfig) -> RawConfig:
    config = cast(RawConfig, _replace(config, lambda x: x == _CONFIG_NONE, None))
    return config


def pack_config(config: RawConfig) -> RawConfig:
    config = cast(RawConfig, _replace(config, lambda x: x is None, _CONFIG_NONE))
    return config


def load_config(path: Union[Path, str]) -> Any:
    with open(path, 'rb') as f:
        return unpack_config(tomli.load(f))


def dump_config(config: Any, path: Union[Path, str]) -> None:
    with open(path, 'wb') as f:
        tomli_w.dump(pack_config(config), f)
    # check that there are no bugs in all these "pack/unpack" things
    assert config == load_config(path)


def load_json(path: Union[Path, str], **kwargs) -> Any:
    return json.loads(Path(path).read_text(), **kwargs)


def dump_json(x: Any, path: Union[Path, str], **kwargs) -> None:
    kwargs.setdefault('indent', 4)
    Path(path).write_text(json.dumps(x, **kwargs) + '\n')


def load_pickle(path: Union[Path, str], **kwargs) -> Any:
    return pickle.loads(Path(path).read_bytes(), **kwargs)


def dump_pickle(x: Any, path: Union[Path, str], **kwargs) -> None:
    Path(path).write_bytes(pickle.dumps(x, **kwargs))


def load(path: Union[Path, str], **kwargs) -> Any:
    return globals()[f'load_{Path(path).suffix[1:]}'](Path(path), **kwargs)


def dump(x: Any, path: Union[Path, str], **kwargs) -> Any:
    return globals()[f'dump_{Path(path).suffix[1:]}'](x, Path(path), **kwargs)


def _get_output_item_path(
        path: Union[str, Path], filename: str, must_exist: bool
) -> Path:
    path = env.get_path(path)
    if path.suffix == '.toml':
        path = path.with_suffix('')
    if path.is_dir():
        path = path / filename
    else:
        assert path.name == filename
    assert path.parent.exists()
    if must_exist:
        assert path.exists()
    return path


def load_report(path: Path) -> Report:
    return load_json(_get_output_item_path(path, 'report.json', True))


def dump_report(report: dict, path: Path) -> None:
    dump_json(report, _get_output_item_path(path, 'report.json', False))


def load_predictions(path: Path) -> Dict[str, np.ndarray]:
    with np.load(_get_output_item_path(path, 'predictions.npz', True)) as predictions:
        return {x: predictions[x] for x in predictions}


def dump_predictions(predictions: Dict[str, np.ndarray], path: Path) -> None:
    np.savez(_get_output_item_path(path, 'predictions.npz', False), **predictions)


def dump_metrics(metrics: Dict[str, Any], path: Path) -> None:
    dump_json(metrics, _get_output_item_path(path, 'metrics.json', False))


def load_checkpoint(path: Path, *args, **kwargs) -> Dict[str, np.ndarray]:
    return torch.load(
        _get_output_item_path(path, 'checkpoint.pt', True), *args, **kwargs
    )


def get_device() -> torch.device:
    if torch.cuda.is_available():
        assert os.environ.get('CUDA_VISIBLE_DEVICES') is not None
        return torch.device('cuda:0')
    else:
        return torch.device('cpu')


def _print_sep(c, size=100):
    print(c * size)


def start(
        config_cls: Type[T] = RawConfig,
        argv: Optional[List[str]] = None,
        patch_raw_config: Optional[Callable[[RawConfig], None]] = None,
) -> Tuple[T, Path, Report]:  # config  # output dir  # report
    parser = argparse.ArgumentParser()
    parser.add_argument('config', metavar='FILE')
    parser.add_argument('--force', action='store_true')
    parser.add_argument('--continue', action='store_true', dest='continue_')
    if argv is None:
        program = __main__.__file__
        args = parser.parse_args()
    else:
        program = argv[0]
        try:
            args = parser.parse_args(argv[1:])
        except Exception:
            print(
                'Failed to parse `argv`.'
                ' Remember that the first item of `argv` must be the path (relative to'
                ' the project root) to the script/notebook.'
            )
            raise
    args = parser.parse_args(argv)

    snapshot_dir = os.environ.get('SNAPSHOT_PATH')
    if snapshot_dir and Path(snapshot_dir).joinpath('CHECKPOINTS_RESTORED').exists():
        assert args.continue_

    config_path = env.get_path(args.config)
    output_dir = config_path.with_suffix('')
    _print_sep('=')
    print(f'[output] {output_dir}')
    _print_sep('=')

    assert config_path.exists()
    raw_config = load_config(config_path)
    if patch_raw_config is not None:
        patch_raw_config(raw_config)
    if is_dataclass(config_cls):
        config = from_dict(config_cls, raw_config)
        full_raw_config = asdict(config)
    else:
        assert config_cls is dict
        full_raw_config = config = raw_config
    full_raw_config = asdict(config)

    if output_dir.exists():
        if args.force:
            print('Removing the existing output and creating a new one...')
            shutil.rmtree(output_dir)
            output_dir.mkdir()
        elif not args.continue_:
            backup_output(output_dir)
            print('The output directory already exists. Done!\n')
            sys.exit()
        elif output_dir.joinpath('DONE').exists():
            backup_output(output_dir)
            print('The "DONE" file already exists. Done!')
            sys.exit()
        else:
            print('Continuing with the existing output...')
    else:
        print('Creating the output...')
        output_dir.mkdir()

    report = {
        'program': str(env.get_relative_path(program)),
        'environment': {},
        'config': full_raw_config,
    }
    if torch.cuda.is_available():  # type: ignore[code]
        report['environment'].update(
            {
                'CUDA_VISIBLE_DEVICES': os.environ.get('CUDA_VISIBLE_DEVICES'),
                'gpus': zero.hardware.get_gpus_info(),
                'torch.version.cuda': torch.version.cuda,
                'torch.backends.cudnn.version()': torch.backends.cudnn.version(),  # type: ignore[code]
                'torch.cuda.nccl.version()': torch.cuda.nccl.version(),  # type: ignore[code]
            }
        )
    dump_report(report, output_dir)
    dump_json(raw_config, output_dir / 'raw_config.json')
    _print_sep('-')
    pprint(full_raw_config, width=100)
    _print_sep('-')
    return cast(config_cls, config), output_dir, report


_LAST_SNAPSHOT_TIME = None


def backup_output(output_dir: Path) -> None:
    backup_dir = os.environ.get('TMP_OUTPUT_PATH')
    snapshot_dir = os.environ.get('SNAPSHOT_PATH')
    if backup_dir is None:
        assert snapshot_dir is None
        return
    assert snapshot_dir is not None

    try:
        relative_output_dir = output_dir.relative_to(env.PROJ)
    except ValueError:
        return

    for dir_ in [backup_dir, snapshot_dir]:
        new_output_dir = dir_ / relative_output_dir
        prev_backup_output_dir = new_output_dir.with_name(new_output_dir.name + '_prev')
        new_output_dir.parent.mkdir(exist_ok=True, parents=True)
        if new_output_dir.exists():
            new_output_dir.rename(prev_backup_output_dir)
        shutil.copytree(output_dir, new_output_dir)
        # the case for evaluate.py which automatically creates configs
        if output_dir.with_suffix('.toml').exists():
            shutil.copyfile(
                output_dir.with_suffix('.toml'), new_output_dir.with_suffix('.toml')
            )
        if prev_backup_output_dir.exists():
            shutil.rmtree(prev_backup_output_dir)

    global _LAST_SNAPSHOT_TIME
    if _LAST_SNAPSHOT_TIME is None or time.time() - _LAST_SNAPSHOT_TIME > 10 * 60:
        import nirvana_dl.snapshot  # type: ignore[code]

        nirvana_dl.snapshot.dump_snapshot()
        _LAST_SNAPSHOT_TIME = time.time()
        print('The snapshot was saved!')


def _get_scores(metrics: Dict[str, Dict[str, Any]]) -> Optional[Dict[str, float]]:
    return (
        {k: v['score'] for k, v in metrics.items()}
        if 'score' in next(iter(metrics.values()))
        else None
    )


def format_scores(metrics: Dict[str, Dict[str, Any]]) -> str:
    return ' '.join(
        f"[{x}] {metrics[x]['score']:.3f}"
        for x in ['test', 'val', 'train']
        if x in metrics
    )


def finish(output_dir: Path, report: dict) -> None:
    print()
    _print_sep('=')

    metrics = report.get('metrics')
    if metrics is not None:
        scores = _get_scores(metrics)
        if scores is not None:
            dump_json(scores, output_dir / 'scores.json')
            print(format_scores(metrics))
            _print_sep('-')

    dump_report(report, output_dir)
    json_output_path = os.environ.get('JSON_OUTPUT_FILE')
    if json_output_path:
        try:
            key = str(output_dir.relative_to(env.PROJ))
        except ValueError:
            pass
        else:
            json_output_path = Path(json_output_path)
            try:
                json_data = json.loads(json_output_path.read_text())
            except (FileNotFoundError, json.decoder.JSONDecodeError):
                json_data = {}
            json_data[key] = load_json(output_dir / 'report.json')
            json_output_path.write_text(json.dumps(json_data, indent=4))
        shutil.copyfile(
            json_output_path,
            os.path.join(os.environ['SNAPSHOT_PATH'], 'json_output.json'),
        )

    output_dir.joinpath('DONE').touch()
    backup_output(output_dir)
    print(f'Done! | {report.get("time")} | {output_dir}')
    _print_sep('=')
    print()


def from_dict(datacls: Type[T], data: dict) -> T:
    assert is_dataclass(datacls)
    data = deepcopy(data)
    for field in fields(datacls):
        if field.name not in data:
            continue
        if is_dataclass(field.type):
            data[field.name] = from_dict(field.type, data[field.name])
        elif (
                get_origin(field.type) is Union
                and len(get_args(field.type)) == 2
                and get_args(field.type)[1] is type(None)
                and is_dataclass(get_args(field.type)[0])
        ):
            if data[field.name] is not None:
                data[field.name] = from_dict(get_args(field.type)[0], data[field.name])
    return datacls(**data)


def replace_factor_with_value(
        config: RawConfig,
        key: str,
        reference_value: int,
        bounds: Tuple[float, float],
) -> None:
    factor_key = key + '_factor'
    if factor_key not in config:
        assert key in config
    else:
        assert key not in config
        factor = config.pop(factor_key)
        assert bounds[0] <= factor <= bounds[1]
        config[key] = int(factor * reference_value)


def get_temporary_copy(path: Union[str, Path]) -> Path:
    path = env.get_path(path)
    assert not path.is_dir() and not path.is_symlink()
    tmp_path = path.with_name(
        path.stem + '___' + str(uuid.uuid4()).replace('-', '') + path.suffix
    )
    shutil.copyfile(path, tmp_path)
    atexit.register(lambda: tmp_path.unlink())
    return tmp_path


def get_python():
    python = Path('python3.9')
    return str(python) if python.exists() else 'python'


def get_catboost_config(real_data_path, is_cv=False):
    ds_name = Path(real_data_path).name
    C = load_json(f'tuned_models/catboost/{ds_name}_cv.json')
    return C



from .model import *


class sep_MLP(nn.Module):
    def __init__(self, dim, len_feats, categories):
        super(sep_MLP, self).__init__()
        self.len_feats = len_feats
        self.layers = nn.ModuleList([])
        for i in range(len_feats):
            self.layers.append(simple_MLP([dim, 5 * dim, categories[i]]))

    def forward(self, x):
        y_pred = list([])
        for i in range(self.len_feats):
            x_i = x[:, i, :]
            pred = self.layers[i](x_i)
            y_pred.append(pred)
        return y_pred


class SAINT(nn.Module):
    def __init__(
            self,
            *,
            categories,
            num_continuous,
            dim,
            depth,
            heads,
            dim_head=16,
            dim_out=1,
            mlp_hidden_mults=(4, 2),
            mlp_act=None,
            num_special_tokens=0,
            attn_dropout=0.,
            ff_dropout=0.,
            cont_embeddings='MLP',
            scalingfactor=10,
            attentiontype='col',
            final_mlp_style='common',
            y_dim=2
    ):
        super().__init__()
        assert all(map(lambda n: n > 0, categories)), 'number of each category must be positive'

        # categories related calculations

        self.num_categories = len(categories)
        self.num_unique_categories = sum(categories)

        # create category embeddings table

        self.num_special_tokens = num_special_tokens
        self.total_tokens = self.num_unique_categories + num_special_tokens

        # for automatically offsetting unique category ids to the correct position in the categories embedding table

        categories_offset = F.pad(torch.tensor(list(categories)), (1, 0), value=num_special_tokens)
        categories_offset = categories_offset.cumsum(dim=-1)[:-1]

        self.register_buffer('categories_offset', categories_offset)

        self.norm = nn.LayerNorm(num_continuous)
        self.num_continuous = num_continuous
        self.dim = dim
        self.cont_embeddings = cont_embeddings
        self.attentiontype = attentiontype
        self.final_mlp_style = final_mlp_style

        if self.cont_embeddings == 'MLP':
            self.simple_MLP = nn.ModuleList([simple_MLP([1, 100, self.dim]) for _ in range(self.num_continuous)])
            input_size = (dim * self.num_categories) + (dim * num_continuous)
            nfeats = self.num_categories + num_continuous
        elif self.cont_embeddings == 'pos_singleMLP':
            self.simple_MLP = nn.ModuleList([simple_MLP([1, 100, self.dim]) for _ in range(1)])
            input_size = (dim * self.num_categories) + (dim * num_continuous)
            nfeats = self.num_categories + num_continuous
        else:
            print('Continous features are not passed through attention')
            input_size = (dim * self.num_categories) + num_continuous
            nfeats = self.num_categories

            # transformer
        if attentiontype == 'col':
            self.transformer = Transformer(
                num_tokens=self.total_tokens,
                dim=dim,
                depth=depth,
                heads=heads,
                dim_head=dim_head,
                attn_dropout=attn_dropout,
                ff_dropout=ff_dropout
            )
        elif attentiontype in ['row', 'colrow']:
            self.transformer = RowColTransformer(
                num_tokens=self.total_tokens,
                dim=dim,
                nfeats=nfeats,
                depth=depth,
                heads=heads,
                dim_head=dim_head,
                attn_dropout=attn_dropout,
                ff_dropout=ff_dropout,
                style=attentiontype
            )

        l = input_size // 8
        hidden_dimensions = list(map(lambda t: l * t, mlp_hidden_mults))
        all_dimensions = [input_size, *hidden_dimensions, dim_out]

        self.mlp = MLP(all_dimensions, act=mlp_act)
        self.embeds = nn.Embedding(self.total_tokens, self.dim)  # .to(device)

        cat_mask_offset = F.pad(torch.Tensor(self.num_categories).fill_(2).type(torch.int8), (1, 0), value=0)
        cat_mask_offset = cat_mask_offset.cumsum(dim=-1)[:-1]

        con_mask_offset = F.pad(torch.Tensor(self.num_continuous).fill_(2).type(torch.int8), (1, 0), value=0)
        con_mask_offset = con_mask_offset.cumsum(dim=-1)[:-1]

        self.register_buffer('cat_mask_offset', cat_mask_offset)
        self.register_buffer('con_mask_offset', con_mask_offset)

        self.mask_embeds_cat = nn.Embedding(self.num_categories * 2, self.dim)
        self.mask_embeds_cont = nn.Embedding(self.num_continuous * 2, self.dim)
        self.single_mask = nn.Embedding(2, self.dim)
        self.pos_encodings = nn.Embedding(self.num_categories + self.num_continuous, self.dim)

        if self.final_mlp_style == 'common':
            self.mlp1 = simple_MLP([dim, (self.total_tokens) * 2, self.total_tokens])
            self.mlp2 = simple_MLP([dim, (self.num_continuous), 1])

        else:
            self.mlp1 = sep_MLP(dim, self.num_categories, categories)
            self.mlp2 = sep_MLP(dim, self.num_continuous, np.ones(self.num_continuous).astype(int))

        self.mlpfory = simple_MLP([dim, 1000, y_dim])
        self.pt_mlp = simple_MLP([dim * (self.num_continuous + self.num_categories),
                                  6 * dim * (self.num_continuous + self.num_categories) // 5,
                                  dim * (self.num_continuous + self.num_categories) // 2])
        self.pt_mlp2 = simple_MLP([dim * (self.num_continuous + self.num_categories),
                                   6 * dim * (self.num_continuous + self.num_categories) // 5,
                                   dim * (self.num_continuous + self.num_categories) // 2])

    def forward(self, x_categ, x_cont):

        x = self.transformer(x_categ, x_cont)
        cat_outs = self.mlp1(x[:, :self.num_categories, :])
        con_outs = self.mlp2(x[:, self.num_categories:, :])
        return cat_outs, con_outs
